# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import gc
from time import sleep

from google.protobuf.json_format import MessageToDict
import pytest
from util import create_datamodel_root_in_server, create_root_using_datamodelgen

from ansys.api.fluent.v0 import datamodel_se_pb2
from ansys.api.fluent.v0.variant_pb2 import Variant
import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples
from ansys.fluent.core.services.datamodel_se import (
    PyCommand,
    PyMenuGeneric,
    PyNumerical,
    PySingletonCommandArgumentsSubItem,
    PyTextualCommandArgumentsSubItem,
    ReadOnlyObjectError,
    _convert_value_to_variant,
    _convert_variant_to_value,
    convert_path_to_se_path,
)
from ansys.fluent.core.streaming_services.datamodel_streaming import DatamodelStream
from ansys.fluent.core.utils.execution import timeout_loop
from ansys.fluent.core.utils.fluent_version import FluentVersion


@pytest.mark.parametrize(
    "value,expected",
    [
        (None, None),
        (False, False),
        (True, True),
        (1, 1),
        (1.0, 1.0),
        ("abc", "abc"),
        ((1, 2, 3), [1, 2, 3]),
        ([1, 2, 3], [1, 2, 3]),
        ({"a": 5}, {"a": 5}),
        ({"a": [1, 2, 3]}, {"a": [1, 2, 3]}),
        ({"a": {"b": 5}}, {"a": {"b": 5}}),
    ],
)
def test_convert_value_to_variant_to_value(value, expected):
    variant = Variant()
    _convert_value_to_variant(value, variant)
    assert expected == _convert_variant_to_value(variant)


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_event_subscription(new_meshing_session):
    session = new_meshing_session
    session.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    tags = [
        "/workflow/created/TaskObject",
        "/workflow/modified/TaskObject:TaskObject1/Arguments",
        "/workflow/deleted/TaskObject:TaskObject1",
        "/workflow/affected/TaskObject:TaskObject1",
        "/workflow/affected/TaskObject",
        "/workflow/attribute_changed/TaskObject:TaskObject1/TaskList/isActive",
        "/workflow/command_attribute_changed/InitializeWorkflow/arguments",
    ]
    version = session.get_fluent_version()
    if version < FluentVersion.v252:
        tags.append("/workflow/command_executed/InitializeWorkflow")
    else:
        # TODO: path should be appended to the tag
        tags.append("/workflow/command_executed")
    request = datamodel_se_pb2.SubscribeEventsRequest()
    e1 = request.eventrequest.add(rules="workflow")
    e1.createdEventRequest.parentpath = ""
    e1.createdEventRequest.childtype = "TaskObject"
    e2 = request.eventrequest.add(rules="workflow")
    e2.modifiedEventRequest.path = "TaskObject:TaskObject1/Arguments"
    e3 = request.eventrequest.add(rules="workflow")
    e3.deletedEventRequest.path = "TaskObject:TaskObject1"
    e4 = request.eventrequest.add(rules="workflow")
    e4.affectedEventRequest.path = "TaskObject:TaskObject1"
    e5 = request.eventrequest.add(rules="workflow")
    e5.affectedEventRequest.path = ""
    e5.affectedEventRequest.subtype = "TaskObject"
    e6 = request.eventrequest.add(rules="workflow")
    e6.attributeChangedEventRequest.path = "TaskObject:TaskObject1/TaskList"
    e6.attributeChangedEventRequest.attribute = "isActive"
    e7 = request.eventrequest.add(rules="workflow")
    e7.commandAttributeChangedEventRequest.path = ""
    e7.commandAttributeChangedEventRequest.command = "InitializeWorkflow"
    e7.commandAttributeChangedEventRequest.attribute = "arguments"
    e8 = request.eventrequest.add(rules="workflow")
    e8.commandExecutedEventRequest.path = ""
    e8.commandExecutedEventRequest.command = "InitializeWorkflow"
    response = session._datamodel_service_se.subscribe_events(MessageToDict(request))
    assert all(
        [
            r["status"] == datamodel_se_pb2.STATUS_SUBSCRIBED and r["tag"] == t
            for r, t in zip(response, tags)
        ]
    )
    response = session._datamodel_service_se.unsubscribe_events(tags)
    assert all(
        [
            r["status"] == datamodel_se_pb2.STATUS_UNSUBSCRIBED and r["tag"] == t
            for r, t in zip(response, tags)
        ]
    )


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_add_on_child_created(new_meshing_session):
    meshing = new_meshing_session
    child_paths = []
    subscription = meshing.workflow.add_on_child_created(
        "TaskObject", lambda obj: child_paths.append(convert_path_to_se_path(obj.path))
    )
    assert child_paths == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    sleep(5)
    assert len(child_paths) > 0
    child_paths.clear()
    subscription.unsubscribe()
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert child_paths == []


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_add_on_deleted(new_meshing_session):
    meshing = new_meshing_session
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    data = []
    _ = meshing.workflow.TaskObject["Import Geometry"].add_on_deleted(
        lambda: data.append(True)
    )
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert len(data) > 0


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_add_on_changed(new_meshing_session):
    meshing = new_meshing_session
    task_list = meshing.workflow.Workflow.TaskList
    assert isinstance(task_list(), list)
    assert len(task_list()) == 0
    data = []
    subscription = task_list.add_on_changed(lambda obj: data.append(True))
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    sleep(5)
    assert len(data) > 0
    assert len(task_list()) > 0
    data.clear()
    subscription.unsubscribe()
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert data == []


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_add_on_affected(new_meshing_session):
    meshing = new_meshing_session
    data = []
    subscription = meshing.workflow.Workflow.add_on_affected(
        lambda obj: data.append(True)
    )
    assert data == []
    wt = meshing.watertight()
    sleep(5)
    assert len(data) > 0
    assert data[0]

    calls = []
    subscription2 = meshing.workflow.add_on_affected(lambda obj: calls.append(True))
    geom = examples.download_file(
        file_name="mixing_elbow.pmdb", directory="pyfluent/mixing_elbow"
    )
    import_geom = meshing.workflow.TaskObject["Import Geometry"]
    assert "FileName" not in import_geom.Arguments()
    assert wt.import_geometry.command_arguments()["FileName"] is None
    import_geom.Arguments = {"FileName": geom}
    assert import_geom.Arguments()["FileName"] == geom
    assert wt.import_geometry.command_arguments()["FileName"] == geom
    sleep(1)
    assert calls == [True]
    import_geom.Arguments = {"FileName": "dummy"}
    sleep(1)
    assert calls == 2 * [True]
    import_geom.Arguments = {"FileName": geom}
    sleep(1)
    assert calls == 3 * [True]
    execute_state = [meshing.workflow()]
    import_geom.Execute()
    calls_after_execute = []
    loop_count = 10
    for i in range(loop_count):
        sleep(5)
        calls_after_execute.append(list(calls))
        execute_state.append(meshing.workflow())
    assert all(state == execute_state[1] for state in execute_state[2:])
    call_count = sum(map(len, calls_after_execute))
    assert call_count == 6 * loop_count
    assert calls_after_execute == loop_count * [6 * [True]]
    subscription2.unsubscribe()

    data.clear()
    subscription.unsubscribe()
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert data == []


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_add_on_affected_at_type_path(new_meshing_session):
    meshing = new_meshing_session
    data = []
    subscription = meshing.workflow.add_on_affected_at_type_path(
        "TaskObject", lambda obj: data.append(True)
    )
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    sleep(5)
    assert len(data) > 0
    assert data[0]
    data.clear()
    subscription.unsubscribe()
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert data == []


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_add_on_command_executed(new_meshing_session):
    meshing = new_meshing_session
    data = []
    version = meshing.get_fluent_version()
    if version < FluentVersion.v252:
        subscription = meshing.meshing.add_on_command_executed_old(
            "ImportGeometry", lambda obj, command, args: data.append(True)
        )
    else:
        subscription = meshing.meshing.add_on_command_executed(
            lambda obj, command, args: data.append(True)
        )
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    meshing.meshing.ImportGeometry(FileName=import_file_name)
    sleep(5)
    assert len(data) > 0
    assert data[0]
    data.clear()
    subscription.unsubscribe()
    meshing.meshing.ImportGeometry(FileName=import_file_name)
    sleep(5)
    assert data == []


@pytest.mark.skip("https://github.com/ansys/pyfluent/issues/2999")
@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_datamodel_streaming_full_diff_state(
    disable_datamodel_cache, new_meshing_session
):
    meshing = new_meshing_session
    datamodel_service_se = meshing._datamodel_service_se
    stream = DatamodelStream(datamodel_service_se)
    stream.start(rules="meshing", no_commands_diff_state=False)

    def cb(state, deleted_paths):
        if state:
            state = _convert_variant_to_value(state)
        cb.states.append(state)

    cb.states = []

    stream.register_callback(cb)

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    meshing.meshing.ImportGeometry(FileName=import_file_name)
    sleep(5)
    assert "ImportGeometry:ImportGeometry1" in (y for x in cb.states for y in x)


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_datamodel_streaming_no_commands_diff_state(
    disable_datamodel_cache, new_meshing_session
):
    meshing = new_meshing_session
    datamodel_service_se = meshing._datamodel_service_se
    stream = DatamodelStream(datamodel_service_se)
    stream.start(rules="meshing", no_commands_diff_state=True)

    def cb(state, deleted_paths):
        if state:
            state = _convert_variant_to_value(state)
        cb.states.append(state)

    cb.states = []

    stream.register_callback(cb)

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    meshing.meshing.ImportGeometry(FileName=import_file_name)
    sleep(5)
    assert "ImportGeometry:ImportGeometry1" not in (y for x in cb.states for y in x)


@pytest.mark.fluent_version(">=24.2")
@pytest.mark.codegen_required
def test_get_object_names_wtm(new_meshing_session):
    meshing = new_meshing_session

    assert not meshing.workflow.TaskObject.get_object_names()

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")

    child_object_names = [
        "Import Geometry",
        "Add Local Sizing",
        "Generate the Surface Mesh",
        "Describe Geometry",
        "Apply Share Topology",
        "Enclose Fluid Regions (Capping)",
        "Update Boundaries",
        "Create Regions",
        "Update Regions",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    ]

    assert meshing.workflow.TaskObject.get_object_names() == child_object_names


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_get_and_set_state_for_command_arg_instance(new_meshing_session):
    meshing = new_meshing_session

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")

    x = meshing.meshing.ImportGeometry.create_instance()

    assert x.LengthUnit() == "mm"

    assert x.LengthUnit.allowed_values() == ["m", "cm", "mm", "in", "ft", "um", "nm"]

    x.LengthUnit.set_state("ft")

    assert x.LengthUnit.get_state() == "ft"

    assert x.CadImportOptions.ExtractFeatures()

    x.CadImportOptions.ExtractFeatures.set_state(False)

    assert not x.CadImportOptions.ExtractFeatures()

    x.set_state({"FileName": "dummy_file_name.dummy_extn"})

    assert x.FileName() == "dummy_file_name.dummy_extn"


def _is_internal_name(name: str, prefix: str) -> bool:
    return name.startswith(prefix) and name.removeprefix(prefix).isdigit()


@pytest.mark.codegen_required
def test_task_object_keys_are_display_names(new_meshing_session):
    meshing = new_meshing_session
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    task_object_state = meshing.workflow.TaskObject()
    assert len(task_object_state) > 0
    assert not any(_is_internal_name(x, "TaskObject:") for x in task_object_state)


def test_generic_datamodel(new_solver_session):
    solver = new_solver_session
    solver.scheme.eval("(init-flserver)")
    flserver = PyMenuGeneric(solver._datamodel_service_se, "flserver")
    assert flserver.Case.Solution.Calculation.TimeStepSize() == 1.0


@pytest.mark.fluent_version(">=24.2")
def test_named_object_specific_methods_using_flserver(new_solver_session):
    import_file_name = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )
    solver = new_solver_session
    solver.file.read(file_type="case", file_name=import_file_name)
    solver.solution.initialization.hybrid_initialize()
    solver.solution.run_calculation.iterate(iter_count=10)
    solver.tui.display.objects.create(
        "contour",
        "contour-z1",
        "field",
        "velocity-magnitude",
        "surfaces-list",
        "cold-inlet",
    )
    solver.tui.display.objects.create(
        "contour",
        "contour-z2",
        "field",
        "velocity-magnitude",
        "surfaces-list",
        "hot-inlet",
    )
    solver.tui.display.objects.create(
        "contour",
        "contour-z3",
        "field",
        "velocity-magnitude",
        "surfaces-list",
        "outlet",
    )
    solver.tui.display.objects.create(
        "contour",
        "contour-z4",
        "field",
        "velocity-magnitude",
        "surfaces-list",
        "wall-elbow",
    )
    solver.tui.display.objects.create(
        "contour",
        "contour-z5",
        "field",
        "velocity-magnitude",
        "surfaces-list",
        "wall-inlet",
    )

    flserver = PyMenuGeneric(solver._datamodel_service_se, "flserver")

    assert set(flserver.Case.Results.Graphics.Contour.get_object_names()) == {
        "contour-z1",
        "contour-z2",
        "contour-z3",
        "contour-z4",
        "contour-z5",
    }

    assert "contour-x1" not in flserver.Case.Results.Graphics.Contour.get_object_names()

    flserver.Case.Results.Graphics.Contour["contour-z1"].rename("contour-x1")

    assert "contour-x1" in flserver.Case.Results.Graphics.Contour.get_object_names()

    flserver.Case.Results.Graphics.delete_child_objects(
        "Contour", ["contour-x1", "contour-z2"]
    )

    assert set(flserver.Case.Results.Graphics.Contour.get_object_names()) == {
        "contour-z3",
        "contour-z4",
        "contour-z5",
    }

    flserver.Case.Results.Graphics.delete_all_child_objects("Contour")

    assert not flserver.Case.Results.Graphics.Contour.get_object_names()


@pytest.mark.fluent_version(">=24.2")
def test_named_object_specific_methods(new_meshing_session):
    meshing = new_meshing_session
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")

    assert set(meshing.workflow.TaskObject.get_object_names()) == {
        "Import Geometry",
        "Add Local Sizing",
        "Generate the Surface Mesh",
        "Describe Geometry",
        "Apply Share Topology",
        "Enclose Fluid Regions (Capping)",
        "Update Boundaries",
        "Create Regions",
        "Update Regions",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    }

    assert "xyz" not in meshing.workflow.TaskObject.get_object_names()

    meshing.workflow.TaskObject["Import Geometry"].rename("xyz")

    assert "xyz" in meshing.workflow.TaskObject.get_object_names()

    meshing.workflow.delete_all_child_objects("TaskObject")

    assert not meshing.workflow.TaskObject.get_object_names()

    # test availability of different ways to get container state:
    TaskObject = meshing.workflow.TaskObject
    assert TaskObject.get_state() == TaskObject.getState() == TaskObject()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_command_creation_inside_singleton(new_meshing_session):
    meshing = new_meshing_session
    read_mesh = meshing.meshing.File.ReadMesh.create_instance()
    assert read_mesh.FileName is not None


@pytest.mark.codegen_required
def test_read_only_set_state(new_meshing_session):
    meshing = new_meshing_session
    meshing.preferences.MeshingWorkflow.SaveCheckpointFiles = True
    assert meshing.preferences.MeshingWorkflow.CheckpointingOption.is_read_only()
    with pytest.raises(ReadOnlyObjectError):
        meshing.preferences.MeshingWorkflow.CheckpointingOption = "Write into memory"

    assert "set_state" not in dir(
        meshing.preferences.MeshingWorkflow.CheckpointingOption
    )


test_rules = (
    "RULES:\n"
    "    SINGLETON: ROOT\n"
    "        members = A\n"
    "        OBJECT: A\n"
    "            members = B, X\n"
    "            commands = C\n"
    "            isABC = $./X == ABC\n"
    "            OBJECT: B\n"
    "            END\n"
    "            STRING: X\n"
    "                default = IJK\n"
    "            END\n"
    "            COMMAND: C\n"
    "                returnType = Logical\n"
    "                functionName = S_C\n"
    "                isABC = $../X == ABC\n"
    "            END\n"
    "        END\n"
    "     END\n"
    "END\n"
)


@pytest.mark.fluent_version(">=24.2")
def test_on_child_created_lifetime(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].add_on_child_created("B", lambda _: data.append(1))
    root.A["A1"].add_on_child_created("B", lambda _: data.append(2))
    gc.collect()
    assert "/test/created/A:A1/B" in solver._se_service.subscriptions
    assert "/test/created/A:A1/B-1" in solver._se_service.subscriptions
    root.A["A1"].B["B1"] = {}
    assert timeout_loop(lambda: data == [1, 2], 5)
    del root.A["A1"]
    assert "/test/created/A:A1/B" not in solver._se_service.subscriptions
    assert "/test/created/A:A1/B-1" not in solver._se_service.subscriptions


@pytest.mark.fluent_version(">=24.2")
def test_on_deleted_lifetime(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].add_on_deleted(lambda: data.append(1))
    root.A["A1"].add_on_deleted(lambda: data.append(2))
    gc.collect()
    assert "/test/deleted/A:A1" in solver._se_service.subscriptions
    assert "/test/deleted/A:A1-1" in solver._se_service.subscriptions
    del root.A["A1"]
    assert timeout_loop(lambda: data == [1, 2], 5)
    assert timeout_loop(
        lambda: "/test/deleted/A:A1" not in solver._se_service.subscriptions, 5
    )
    assert timeout_loop(
        lambda: "/test/deleted/A:A1-1" not in solver._se_service.subscriptions, 5
    )


@pytest.mark.fluent_version(">=24.2")
def test_on_changed_lifetime(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].X.add_on_changed(lambda _: data.append(1))
    root.A["A1"].X.add_on_changed(lambda _: data.append(2))
    gc.collect()
    assert "/test/modified/A:A1/X" in solver._se_service.subscriptions
    assert "/test/modified/A:A1/X-1" in solver._se_service.subscriptions
    root.A["A1"].X = "ABC"
    assert timeout_loop(lambda: data == [1, 2], 5)
    del root.A["A1"]
    assert "/test/modified/A:A1/X" not in solver._se_service.subscriptions
    assert "/test/modified/A:A1/X-1" not in solver._se_service.subscriptions


@pytest.mark.fluent_version(">=24.2")
def test_on_affected_lifetime(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].add_on_affected(lambda _: data.append(1))
    root.A["A1"].add_on_affected(lambda _: data.append(2))
    gc.collect()
    assert "/test/affected/A:A1" in solver._se_service.subscriptions
    assert "/test/affected/A:A1-1" in solver._se_service.subscriptions
    root.A["A1"].X = "ABC"
    assert timeout_loop(lambda: data == [1, 2], 5)
    del root.A["A1"]
    assert "/test/affected/A:A1" not in solver._se_service.subscriptions
    assert "/test/affected/A:A1-1" not in solver._se_service.subscriptions


@pytest.mark.fluent_version(">=24.2")
def test_on_affected_at_type_path_lifetime(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].add_on_affected_at_type_path("B", lambda _: data.append(1))
    root.A["A1"].add_on_affected_at_type_path("B", lambda _: data.append(2))
    gc.collect()
    assert "/test/affected/A:A1/B" in solver._se_service.subscriptions
    assert "/test/affected/A:A1/B-1" in solver._se_service.subscriptions
    root.A["A1"].B["B1"] = {}
    assert timeout_loop(lambda: data == [1, 2], 5)
    del root.A["A1"]
    assert "/test/affected/A:A1/B" not in solver._se_service.subscriptions
    assert "/test/affected/A:A1/B-1" not in solver._se_service.subscriptions


@pytest.mark.fluent_version(">=24.2")
def test_on_command_executed_lifetime(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    root.A["A1"] = {}
    data = []
    version = solver.get_fluent_version()
    if version < FluentVersion.v252:
        _ = root.A["A1"].add_on_command_executed_old("C", lambda *args: data.append(1))
        root.A["A1"].add_on_command_executed_old("C", lambda *args: data.append(2))
        tags = ["/test/command_executed/A:A1/C", "/test/command_executed/A:A1/C-1"]
    else:
        _ = root.A["A1"].add_on_command_executed(lambda *args: data.append(1))
        root.A["A1"].add_on_command_executed(lambda *args: data.append(2))
        tags = ["/test/command_executed/A:A1", "/test/command_executed/A:A1-1"]
    gc.collect()
    for tag in tags:
        assert tag in solver._se_service.subscriptions
    root.A["A1"].C()
    assert timeout_loop(lambda: data == [1, 2], 5)
    del root.A["A1"]
    for tag in tags:
        assert tag not in solver._se_service.subscriptions


@pytest.mark.fluent_version(">=24.2")
def test_on_attribute_changed_lifetime(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].add_on_attribute_changed("isABC", lambda _: data.append(1))
    root.A["A1"].add_on_attribute_changed("isABC", lambda _: data.append(2))
    gc.collect()
    assert "/test/attribute_changed/A:A1/isABC" in solver._se_service.subscriptions
    assert "/test/attribute_changed/A:A1/isABC-1" in solver._se_service.subscriptions
    root.A["A1"].X = "ABC"
    assert timeout_loop(lambda: data == [1, 2], 5)
    del root.A["A1"]
    assert "/test/attribute_changed/A:A1/isABC" not in solver._se_service.subscriptions
    assert (
        "/test/attribute_changed/A:A1/isABC-1" not in solver._se_service.subscriptions
    )


@pytest.mark.fluent_version(">=24.2")
def test_on_command_attribute_changed_lifetime(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].add_on_command_attribute_changed(
        "C", "isABC", lambda _: data.append(1)
    )
    root.A["A1"].add_on_command_attribute_changed(
        "C", "isABC", lambda _: data.append(2)
    )
    gc.collect()
    assert (
        "/test/command_attribute_changed/A:A1/C/isABC"
        in solver._se_service.subscriptions
    )
    assert (
        "/test/command_attribute_changed/A:A1/C/isABC-1"
        in solver._se_service.subscriptions
    )
    root.A["A1"].X = "ABC"
    assert timeout_loop(lambda: data == [1, 2], 5)
    del root.A["A1"]
    assert (
        "/test/command_attribute_changed/A:A1/C/isABC"
        not in solver._se_service.subscriptions
    )
    assert (
        "/test/command_attribute_changed/A:A1/C/isABC-1"
        not in solver._se_service.subscriptions
    )


@pytest.mark.fluent_version(">=24.2")
def test_on_affected_lifetime_with_delete_child_objects(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    pyfluent.logger.enable()
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].add_on_affected(lambda _: data.append(1))
    root.A["A1"].add_on_affected(lambda _: data.append(2))
    gc.collect()
    assert "/test/affected/A:A1" in solver._se_service.subscriptions
    assert "/test/affected/A:A1-1" in solver._se_service.subscriptions
    root.A["A1"].X = "ABC"
    assert timeout_loop(lambda: data == [1, 2], 5)
    root.delete_child_objects("A", ["A1"])
    assert "/test/affected/A:A1" not in solver._se_service.subscriptions
    assert "/test/affected/A:A1-1" not in solver._se_service.subscriptions


@pytest.mark.fluent_version(">=24.2")
def test_on_affected_lifetime_with_delete_all_child_objects(new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, test_rules, app_name)
    root = create_root_using_datamodelgen(solver._se_service, app_name)
    pyfluent.logger.enable()
    root.A["A1"] = {}
    data = []
    _ = root.A["A1"].add_on_affected(lambda _: data.append(1))
    root.A["A1"].add_on_affected(lambda _: data.append(2))
    gc.collect()
    assert "/test/affected/A:A1" in solver._se_service.subscriptions
    assert "/test/affected/A:A1-1" in solver._se_service.subscriptions
    root.A["A1"].X = "ABC"
    assert timeout_loop(lambda: data == [1, 2], 5)
    root.delete_all_child_objects("A")
    assert "/test/affected/A:A1" not in solver._se_service.subscriptions
    assert "/test/affected/A:A1-1" not in solver._se_service.subscriptions


@pytest.mark.fluent_version(">=23.2")
def test_set_command_args_and_sub_args(new_meshing_session):
    meshing = new_meshing_session
    ig = meshing.meshing.ImportGeometry.create_instance()

    # Command Arguments
    assert ig.MeshUnit() == "m"
    ig.MeshUnit = "mm"
    assert ig.MeshUnit() == "mm"

    # Command Arguments SubItem
    assert ig.CadImportOptions.OneZonePer() == "body"
    ig.CadImportOptions.OneZonePer = "face"
    assert ig.CadImportOptions.OneZonePer() == "face"


@pytest.mark.fluent_version(">=24.1")
def test_dynamic_dependency(new_meshing_session):
    meshing = new_meshing_session
    ic = meshing.meshing.LoadCADGeometry.create_instance()

    d = ic.Refaceting.Deviation.get_state()
    cd = ic.Refaceting.CustomDeviation.get_state()
    assert d == cd

    ic.Refaceting.Deviation.set_state(1.2)
    d = ic.Refaceting.Deviation.get_state()
    cd = ic.Refaceting.CustomDeviation.get_state()
    assert d == cd


@pytest.mark.fluent_version(">=25.2")
def test_field_level_help(new_meshing_session):
    meshing = new_meshing_session
    meshing.PartManagement.AssemblyNode["node-1"] = {}
    deviation = meshing.PartManagement.AssemblyNode["node-1"].Refaceting.Deviation
    assert isinstance(deviation, PyNumerical)
    # Field-level help at parameter level
    assert deviation.__doc__.strip().startswith(
        "Specify the distance between facet edges and the geometry edges. Decreasing this value"
    )
    # TODO Test Field-level help at singleton level when we have that in the datamodel
    assert meshing.meshing.ImportGeometry, PyCommand
    # Field-level help at command level
    assert meshing.meshing.ImportGeometry.__doc__.strip().startswith(
        "Specify the CAD geometry that you want to work with. Choose from"
    )
    import_geometry = meshing.meshing.ImportGeometry.create_instance()
    assert isinstance(import_geometry.FileFormat, PyTextualCommandArgumentsSubItem)
    # Field-level help at parameter-type command argument level
    assert import_geometry.FileFormat.__doc__.strip().startswith(
        "Indicate whether the imported geometry is a CAD File or"
    )
    linear_mesh_pattern = meshing.meshing.LinearMeshPattern.create_instance()
    assert isinstance(
        linear_mesh_pattern.PatternVector, PySingletonCommandArgumentsSubItem
    )
    # Field-level help at singleton-type command argument level
    assert linear_mesh_pattern.PatternVector.__doc__.strip().startswith(
        "Specify a name for the mesh pattern or use the default value."
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Callable

import pytest

import ansys.fluent.core as pyfluent
from ansys.fluent.core.data_model_cache import DataModelCache
from ansys.fluent.core.examples import download_file


def fluent_launcher_args(args: str):
    def fluent_launcher_args_inner(f: Callable):
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)

        wrapper.fluent_launcher_args = args
        return wrapper

    return fluent_launcher_args_inner


def mixing_elbow_geometry_filename(globals):
    return download_file(
        file_name="mixing_elbow.pmdb", directory="pyfluent/mixing_elbow"
    )


def exhaust_system_geometry_filename(globals):
    return download_file(
        file_name="exhaust_system.fmd", directory="pyfluent/exhaust_system"
    )


@fluent_launcher_args("3ddp -meshing")
def new_meshing_session(globals):
    meshing = globals["meshing"]
    return meshing


@fluent_launcher_args("3ddp -meshing")
def new_pure_meshing_session(globals):
    return new_meshing_session(globals)


@fluent_launcher_args("3ddp -meshing")
def watertight_workflow_session(globals):
    meshing = new_meshing_session(globals)
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    return meshing


@fluent_launcher_args("3ddp -meshing")
def fault_tolerant_workflow_session(globals):
    meshing = new_meshing_session(globals)
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    return meshing


@fluent_launcher_args("3ddp -meshing")
def mixing_elbow_watertight_pure_meshing_session(globals):
    meshing = new_pure_meshing_session(globals)
    geometry_filename = mixing_elbow_geometry_filename(globals)
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    meshing.workflow.TaskObject["Import Geometry"].Arguments = dict(
        FileName=geometry_filename, LengthUnit="in"
    )
    return meshing


@fluent_launcher_args("3ddp")
def new_solver_session(globals):
    solver = globals["solver"]
    return solver


@fluent_launcher_args("3d")
def new_solver_session_sp(globals):
    return new_solver_session(globals)


@fluent_launcher_args("2ddp")
def new_solver_session_2d(globals):
    return new_solver_session(globals)


@fluent_launcher_args("3ddp")
def static_mixer_settings_session(globals):
    solver = new_solver_session(globals)
    case_name = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    solver.file.read(
        file_type="case",
        file_name=case_name,
        lightweight_setup=True,
    )
    return solver


@fluent_launcher_args("3ddp")
def static_mixer_case_session(globals):
    solver = new_solver_session(globals)
    case_name = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    solver.file.read(file_type="case", file_name=case_name)
    return solver


@fluent_launcher_args("3ddp")
def mixing_elbow_settings_session(globals):
    solver = new_solver_session(globals)
    case_name = download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
    solver.settings.file.read(
        file_type="case",
        file_name=case_name,
        lightweight_setup=True,
    )
    return solver


@fluent_launcher_args("3ddp")
def mixing_elbow_case_data_session(globals):
    solver = new_solver_session(globals)
    case_name = download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
    download_file("mixing_elbow.dat.h5", "pyfluent/mixing_elbow")
    solver.settings.file.read(file_type="case-data", file_name=case_name)
    return solver


@fluent_launcher_args("3ddp")
def mixing_elbow_param_case_data_session(globals):
    solver = new_solver_session(globals)
    case_name = download_file("elbow_param.cas.h5", "pyfluent/mixing_elbow")
    download_file("elbow_param.dat.h5", "pyfluent/mixing_elbow")
    solver.settings.file.read(file_type="case-data", file_name=case_name)
    return solver


@fluent_launcher_args("2ddp")
def disk_settings_session(globals):
    solver = new_solver_session_2d(globals)
    case_name = download_file("disk.cas.h5", "pyfluent/rotating_disk")
    solver.file.read(
        file_type="case",
        file_name=case_name,
        lightweight_setup=True,
    )
    return solver


@fluent_launcher_args("2ddp")
def disk_case_session(globals):
    solver = new_solver_session_2d(globals)
    case_name = download_file("disk.cas.h5", "pyfluent/rotating_disk")
    solver.file.read(file_type="case", file_name=case_name)
    return solver


@fluent_launcher_args("3ddp")
def periodic_rot_settings_session(globals):
    solver = new_solver_session(globals)
    case_name = download_file(
        "periodic_rot.cas.h5",
        "pyfluent/periodic_rot",
    )
    solver.file.read(
        file_type="case",
        file_name=case_name,
        lightweight_setup=True,
    )
    return solver


monkeypatch = pytest.MonkeyPatch()


def disable_datamodel_cache(globals):
    monkeypatch.setattr(pyfluent, "DATAMODEL_USE_STATE_CACHE", False)


def display_names_as_keys_in_cache(globals):
    DataModelCache.use_display_name = True


def new_meshing_session2(globals):
    session = pyfluent.launch_fluent(mode=pyfluent.LaunchMode.MESHING)
    return session


def new_solver_session2(globals):
    session = pyfluent.launch_fluent()
    return session


def static_mixer_case_session2(globals):
    session = new_solver_session2(globals)
    case_name = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    session.file.read(file_type="case", file_name=case_name)
    return session



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for accessing Fluent units labels per quantity type.

Example
-------
>>> units = get_si_unit_for_fluent_quantity("pressure")

The following code is employed to translate the source data
into Python data structures:

>>> from ansys.fluent.core.filereader import lispy
>>> from ansys.units import Quantity, Unit
>>> from ansys.units.quantity import get_si_value
>>> import re
>>> from pprint import pprint
>>>
>>> fl_unit_re_subs = {
...     'deg': 'radian',
...     'rad': 'radian',
...     'Ohm': 'ohm'
... }
>>>
>>> fl_unit_subs = {'%': ''}
>>>
>>> def replace_units(match):
...     return fl_unit_re_subs[match.group(0)]
...
>>> def substitute_fl_units_with_py_units(fl_units_dict):
...     subs = {}
...     for k, v in fl_units_dict.items():
...         new_val = re.sub('|'.join(r'\b%s\b' % re.escape(s) for s in fl_unit_re_subs), replace_units, v)
...         if new_val != v:
...             subs[k] = new_val
...         else:
...             new_val = fl_unit_subs.get(v, None)
...             if new_val is not None:
...                 subs[k] = new_val
...     print("Substitutions:")
...     for k, v in subs.items():
...         print(f"'{fl_units_dict[k]}' -> '{v}' for '{k}'")
...     fl_units_dict.update(subs)
...     return fl_units_dict
...
>>> def remove_unhandled_units(fl_units_dict):
...     not_si = {}
...     unhandled = {}
...     for k, v in fl_units_dict.items():
...         try:
...             Unit(v)
...             q = Quantity(1.0, v)
...             if get_si_value(q) != 1.0:
...                 not_si[k] = v
...         except Exception:
...             unhandled[k] = v
...     print("Not SI:")
...     pprint(not_si)
...     print("Unhandled:")
...     pprint(unhandled)
...     [fl_units_dict.pop(key) for key in list(not_si) + list(unhandled)]
...     return fl_units_dict
...
>>> def make_python_fl_unit_table(scheme_unit_table):
...     as_list = lispy.parse(scheme_unit_table)[1][1]
...     as_dict = { x[0]:x[1][3].split('"')[1] for x in as_list }
...     return remove_unhandled_units(substitute_fl_units_with_py_units(as_dict))
...

Output from most recent run to generate the table below:

>>> pprint(make_python_fl_unit_table(fl_scheme_unit_table))
Substitutions:
'deg' -> 'radian' for 'angle'
'rad s^-1' -> 'radian s^-1' for 'angular-velocity'
'Ohm m^3' -> 'ohm m^3' for 'contact-resistance-vol'
'deg' -> 'radian' for 'crank-angle'
'Ohm m^2' -> 'ohm m^2' for 'elec-contact-resistance'
'Ohm m' -> 'ohm m' for 'elec-resistivity'
'Ohm' -> 'ohm' for 'elec-resistance'
'%' -> '' for 'percentage'
'N m rad^-1' -> 'N m radian^-1' for 'spring-constant-angular'
Not SI:
{'concentration': 'kmol m^-3',
 'elec-charge': 'A h',
 'molec-wt': 'kg kmol^-1',
 'soot-sitespecies-concentration': 'kmol m^-3'}
Unhandled:
{'crank-angular-velocity': 'rev min^-1',
 'energy-density': 'J/m2',
 'mole-con-henry-const': 'Pa m^3 kgmol^-1',
 'mole-specific-energy': 'J kgmol^-1',
 'mole-specific-entropy': 'J kgmol^-1 K^-1',
 'mole-transfer-rate': 'kgmol m^-3 s^-1',
 'particles-conc': '1.e15-particles/kg',
 'particles-rate': '1.e15 m^-3 s^-1',
 'site-density': 'kgmol m^-2',
 'soot-limiting-nuclei-rate': '1e+15-particles/m3-s',
 'soot-oxidation-constant': 'kg m kgmol^-1 K^-0.5 s^-1',
 'soot-pre-exponential-constant': '1.e15 kg^-1 s^-1',
 'soot-surface-growth-scale-factor': 'kg m kgmol^-1 s^-1',
 'surface-mole-transfer-rate': 'kgmol m^-2 s^-1',
 'univ-gas-constant': 'J K^-1 kgmol^-1',
 'viscosity-consistency-index': 'kg s^n-2 m^-1',
 'wave-length': 'Angstrom'}
"""

from __future__ import annotations

from typing import TypeVar

_fl_unit_table = {
    "acceleration": "m s^-2",
    "angle": "radian",
    "angular-velocity": "radian s^-1",
    "area": "m^2",
    "area-inverse": "m^-2",
    "collision-rate": "m^-3 s^-1",
    "contact-resistance": "m^2 K W^-1",
    "contact-resistance-vol": "ohm m^3",
    "crank-angle": "radian",
    "current": "A",
    "current-density": "A m^-2",
    "current-vol-density": "A m^-3",
    "density": "kg m^-3",
    "density*specific-energy": "J m^-3",
    "density*specific-heat": "J m^-3 K^-1",
    "density*velocity": "kg m^-2 s^-1",
    "density-gradient": "kg m^-4",
    "density-inverse": "m^3 kg^-1",
    "depth": "m",
    "elec-charge-density": "A s m^-3",
    "elec-conductivity": "S m^-1",
    "elec-contact-resistance": "ohm m^2",
    "elec-field": "V m^-1",
    "elec-permittivity": "farad m^-1",
    "elec-resistance": "ohm",
    "elec-resistivity": "ohm m",
    "energy": "J",
    "force": "N",
    "force*time-per-volume": "N s m^-3",
    "force-per-area": "N m^-2",
    "force-per-volume": "N m^-3",
    "frequency": "Hz",
    "gas-constant": "J kg^-1 K^-1",
    "heat-flux": "W m^-2",
    "heat-flux-resolved": "m K s^-1",
    "heat-generation-rate": "W m^-3",
    "heat-transfer-coefficient": "W m^-2 K^-1",
    "ignition-energy": "J mol^-1",
    "kinematic-viscosity": "m^2 s^-1",
    "length": "m",
    "length-inverse": "m^-1",
    "length-time-inverse": "m^-1 s^-1",
    "mag-permeability": "H m^-1",
    "mass": "kg",
    "mass-diffusivity": "m^2 s^-1",
    "mass-flow": "kg s^-1",
    "mass-flow-per-depth": "kg m^-1 s^-1",
    "mass-flow-per-time": "kg s^-2",
    "mass-flux": "kg m^-2 s^-1",
    "mass-transfer-rate": "kg m^-3 s^-1",
    "moment": "N m",
    "moment-of-inertia": "kg m^2",
    "nucleation-rate": "m^-3 s^-1",
    "number-density": "m^-3",
    "percentage": "",
    "power": "W",
    "power-per-time": "W s^-1",
    "pressure": "Pa",
    "pressure-2nd-time-derivative": "Pa s^-2",
    "pressure-gradient": "Pa m^-1",
    "pressure-time-deriv-sqr": "Pa^2 s^-2",
    "pressure-time-derivative": "Pa s^-1",
    "resistance": "m^-1",
    "soot-formation-constant-unit": "kg N^-1 m^-1 s^-1",
    "soot-linear-termination": "m^3 s^-1",
    "source-elliptic-relaxation-function": "kg m^-3 s^-2",
    "source-energy": "W m^-3",
    "source-kinetic-energy": "kg m^-1 s^-3",
    "source-mass": "kg m^-3 s^-1",
    "source-momentum": "N m^-3",
    "source-specific-dissipation-rate": "kg m^-3 s^-2",
    "source-temperature-variance": "K^2 m^-3 s^-1",
    "source-turbulent-dissipation-rate": "kg m^-1 s^-4",
    "source-turbulent-viscosity": "kg m^-1 s^-2",
    "specific-area": "m^2 kg^-1",
    "specific-energy": "J kg^-1",
    "specific-heat": "J kg^-1 K^-1",
    "spring-constant": "N m^-1",
    "spring-constant-angular": "N m radian^-1",
    "stefan-boltzmann-constant": "W m^-2 K^-4",
    "surface-density": "kg m^-2",
    "surface-tension": "N m^-1",
    "surface-tension-gradient": "N m^-1 K^-1",
    "temperature": "K",
    "temperature-difference": "K",
    "temperature-gradient": "K m^-1",
    "temperature-inverse": "K^-1",
    "temperature-variance": "K^2",
    "thermal-conductivity": "W m^-1 K^-1",
    "thermal-resistance": "m^2 K W^-1",
    "thermal-resistivity": "m K W^-1",
    "thermophoretic-diffusivity": "kg m^2 s^-2",
    "time": "s",
    "time-inverse": "s^-1",
    "time-inverse-cubed": "s^-3",
    "time-inverse-squared": "s^-2",
    "turb-kinetic-energy-production": "kg m^-1 s^-3",
    "turbulent-energy-diss-rate": "m^2 s^-3",
    "turbulent-energy-diss-rate-gradient": "m s^-3",
    "turbulent-kinetic-energy": "m^2 s^-2",
    "turbulent-kinetic-energy-gradient": "m s^-2",
    "velocity": "m s^-1",
    "viscosity": "kg m^-1 s^-1",
    "voltage": "V",
    "volume": "m^3",
    "volume-flow-rate": "m^3 s^-1",
    "volume-flow-rate-per-depth": "m^3 s^-1 m^-1",
    "volume-inverse": "m^-3",
    "youngs-modulus": "N m^-2",
}


class InvalidQuantityType(TypeError):
    """Raised on an attempt to get a Quantity with invalid type."""

    def __init__(
        self,
        quantity,
    ) -> None:
        """Initialize InvalidQuantityType."""
        super().__init__(
            f"The specified quantity, '{quantity}' is not a string ({type(quantity)})."
        )


class UnitsNotDefinedForQuantity(ValueError):
    """Raised on an attempt to get undefined units for the specified Quantity."""

    def __init__(
        self,
        quantity: str,
    ) -> None:
        """Initialize UnitsNotDefinedForQuantity."""
        super().__init__(
            f"The units for the specified quantity, '{quantity}' are not defined in PyFluent."
        )


QuantityT = TypeVar("QuantityT")


class UnhandledQuantity(RuntimeError):
    """Raised on an attempt to get an unhandled Quantity."""

    def __init__(
        self,
        path: str,
        quantity: QuantityT,
    ) -> None:
        """Initialize UnhandledQuantity."""
        super().__init__(
            f"Could not handle the quantity, '{quantity}' for the path, {path}."
        )


def get_si_unit_for_fluent_quantity(
    quantity: str | None, unit_table: dict | None = None
):
    """Get the SI unit for the given Fluent quantity.

    Raises
    ------
    InvalidQuantityType
        If ``quantity`` is not a string instance, unless it is None.
    """
    # The settings API should return None for the units-quantity
    # attribute only for dimensionless variables
    if quantity is None:
        return ""
    if not isinstance(quantity, str):
        raise InvalidQuantityType(quantity)
    try:
        return (unit_table or _fl_unit_table)[quantity]
    except KeyError:
        # if it's not configured, None signifies that
        # we don't know the units. no need to raise
        # as this is pretty normal
        pass



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing class encapsulating Fluent connection."""

from typing import Any, Dict

from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.services import SchemeEval
from ansys.fluent.core.session_pure_meshing import PureMeshing
from ansys.fluent.core.session_solver import Solver


class Meshing(PureMeshing):
    """Encapsulates a Fluent meshing session.

    A ``tui`` object
    for meshing TUI commanding, and ``meshing`` and ``workflow``
    objects for access to task-based meshing workflows are all
    exposed here. A ``switch_to_solver`` method is available
    in this mode.
    """

    def __init__(
        self,
        fluent_connection: FluentConnection,
        scheme_eval: SchemeEval,
        file_transfer_service: Any | None = None,
        start_transcript: bool = True,
        launcher_args: Dict[str, Any] | None = None,
    ):
        """Meshing session.

        Parameters
        ----------
        fluent_connection (:ref:`ref_fluent_connection`):
            Encapsulates a Fluent connection.
        scheme_eval: SchemeEval
            Instance of ``SchemeEval`` to execute Fluent's scheme code on.
        file_transfer_service : Optional
            Service for uploading and downloading files.
        start_transcript : bool, optional
            Whether to start the Fluent transcript in the client.
            The default is ``True``, in which case the Fluent
            transcript can be subsequently started and stopped
            using method calls on the ``Session`` object.
        """
        super(Meshing, self).__init__(
            fluent_connection=fluent_connection,
            scheme_eval=scheme_eval,
            file_transfer_service=file_transfer_service,
            start_transcript=start_transcript,
            launcher_args=launcher_args,
        )

    def switch_to_solver(self) -> Any:
        """Switch to solver mode and return a solver session object. Deactivate this
        object's public interface and streaming services.

        Returns
        -------
        Solver
        """
        for cb in self._fluent_connection.finalizer_cbs:
            cb()
        self.tui.switch_to_solution_mode("yes")
        solver_session = Solver(
            fluent_connection=self._fluent_connection,
            scheme_eval=self.scheme,
            file_transfer_service=self._file_transfer_service,
        )
        self._fluent_connection = None
        self.__doc__ = (
            "The meshing session is no longer usable after switching to solution mode."
        )
        return solver_session

    def __getattribute__(self, item: str):
        if super(Meshing, self).__getattribute__(
            "_fluent_connection"
        ) is None and item not in [
            "is_active",
            "_fluent_connection",
        ]:
            raise AttributeError(
                f"'{__class__.__name__}' object has no attribute '{item}'"
            )

        return super(Meshing, self).__getattribute__(item)

    @property
    def tui(self):
        """Meshing TUI root."""
        return super(Meshing, self).tui

    @property
    def meshing(self):
        """Meshing datamodel root."""
        return super(Meshing, self).meshing

    @property
    def meshing_utilities(self):
        """Meshing utilities datamodel root."""
        return super(Meshing, self).meshing_utilities

    @property
    def workflow(self):
        """Workflow datamodel root."""
        return super(Meshing, self).workflow

    @property
    def PartManagement(self):
        """Part management datamodel root."""
        return super(Meshing, self).PartManagement

    @property
    def PMFileManagement(self):
        """Part management file management datamodel root."""
        return super(Meshing, self).PMFileManagement

    @property
    def preferences(self):
        """Preferences datamodel root."""
        return super(Meshing, self).preferences



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module to produce text documentation for a data object."""

import pprint
import pydoc
import sys


def docother(self, object, name=None, mod=None, parent=None, maxlen=None, doc=None):
    """Produce text documentation for a data object."""
    if isinstance(object, list):
        indent_len = len(name and name + " = " or "") + 1
        rep = pprint.pformat(object, width=maxlen, compact=True, indent=indent_len)
        rep = "[" + rep[1:].lstrip()
    else:
        rep = self.repr(object)
        if maxlen:
            line = (name and name + " = " or "") + rep
            chop = maxlen - len(line)
            if chop < 0:
                rep = rep[:chop] + "..."
    line = (name and self.bold(name) + " = " or "") + rep
    #  The source have been changed in 3.9, cpython commit id fbf2786c4c89430e2067016603078cf3500cfe94
    if sys.version_info < (3, 9):
        if doc is not None:
            line += "\n" + self.indent(str(doc))
    else:
        if not doc:
            doc = pydoc.getdoc(object)
        if doc:
            line += "\n" + self.indent(str(doc)) + "\n"
    return line



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Tests for `PhysicalQuantity` objects.
"""

import pytest

import ansys.fluent.core as pf  # noqa: F401
from ansys.fluent.core import examples
from ansys.fluent.core.file_session import FileSession
from ansys.fluent.core.services.field_data import (
    ScalarFieldDataRequest,
    SurfaceDataType,
    SurfaceFieldDataRequest,
    VectorFieldDataRequest,
)
from ansys.units.variable_descriptor import VariableCatalog


def round_off_list_elements(input_list):
    """
    A function to round off list elements.
    """
    for index, value in enumerate(input_list):
        input_list[index] = round(value, 6)

    return input_list


@pytest.mark.fluent_version(">=24.1")
@pytest.mark.developer_only
def test_use_variable_catalog(new_solver_session) -> None:
    """
    A test of `PhysicalQuantity` objects.
    """
    solver = new_solver_session
    settings = solver.settings

    case_name = examples.download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
    settings.file.read(file_type="case", file_name=case_name)

    settings.solution.initialization.hybrid_initialize()

    fields = solver.fields

    temperature = VariableCatalog.TEMPERATURE
    locations = ["hot-inlet"]

    temperature_field_data = fields.field_data.get_field_data(
        ScalarFieldDataRequest(
            field_name=VariableCatalog.TEMPERATURE, surfaces=locations
        )
    )
    assert round(temperature_field_data[locations[0]][0]) == 305

    temperature_min = fields.reduction.minimum(
        expression=temperature, locations=locations
    )
    assert round(temperature_min) == 313

    temperature_solution_data = fields.solution_variable_data.get_data(
        solution_variable_name=temperature, zone_names=locations
    )
    assert round(temperature_solution_data[locations[0]][0]) == 313

    report_defs = settings.solution.report_definitions
    report_defs.surface["yyy"] = {}
    surface = report_defs.surface["yyy"]
    surface.report_type = "surface-areaavg"
    surface.field = temperature
    surface.surface_names = locations

    result = report_defs.compute(report_defs=["yyy"])
    assert round(result[0]["yyy"][0]) == 313


@pytest.mark.developer_only
def test_use_variable_catalog_offline():
    """
    A test of `PhysicalQuantity` objects for offline data.
    """
    case_file_name = examples.download_file(
        "elbow1.cas.h5", "pyfluent/file_session", return_without_path=False
    )
    data_file_name = examples.download_file(
        "elbow1.dat.h5", "pyfluent/file_session", return_without_path=False
    )
    file_session = FileSession()

    # backward compatibility check
    assert file_session.fields.field_data == file_session.field_data

    file_session.read_case(case_file_name)
    file_session.read_data(data_file_name)

    assert round_off_list_elements(
        file_session.field_info.get_scalar_field_range(VariableCatalog.PRESSURE)
    ) == [-339.203452, 339.417934]
    assert len(file_session.field_info.get_scalar_fields_info()) == 29
    assert list(file_session.field_info.get_surfaces_info().keys()) == [
        "wall",
        "symmetry",
        "pressure-outlet-7",
        "velocity-inlet-6",
        "velocity-inlet-5",
        "default-interior",
    ]
    sv_t_wall_request = ScalarFieldDataRequest(
        field_name=VariableCatalog.TEMPERATURE, surfaces=["wall"]
    )
    sv_t_wall = file_session.fields.field_data.get_field_data(sv_t_wall_request)
    assert sv_t_wall["wall"].shape == (3630,)
    assert round(sv_t_wall["wall"][1800], 4) == 313.15

    surface_data = file_session.fields.field_data.get_field_data
    surface_data_wall_request = SurfaceFieldDataRequest(
        data_types=[SurfaceDataType.Vertices], surfaces=[3]
    )
    surface_data_wall = surface_data(surface_data_wall_request)
    assert surface_data_wall[3].shape == (3810, 3)
    assert round(surface_data_wall[3][1500][0], 5) == 0.12406
    assert round(surface_data_wall[3][1500][1], 5) == 0.09525
    assert round(surface_data_wall[3][1500][2], 5) == 0.04216

    surface_data_symmetry_request = SurfaceFieldDataRequest(
        data_types=[SurfaceDataType.FacesConnectivity], surfaces=["symmetry"]
    )
    surface_data_symmetry = surface_data(surface_data_symmetry_request)
    assert len(surface_data_symmetry["symmetry"]) == 2018
    assert list(surface_data_symmetry["symmetry"][1000]) == [1259, 1260, 1227, 1226]

    vector_data = file_session.fields.field_data.get_field_data
    vector_data_request = VectorFieldDataRequest(
        field_name=VariableCatalog.VELOCITY, surfaces=["wall"]
    )
    assert vector_data(vector_data_request)["wall"].shape == (3630, 3)

    vector_data_symmetry_request = VectorFieldDataRequest(
        field_name=VariableCatalog.VELOCITY, surfaces=["symmetry"]
    )
    vector_data_symmetry = vector_data(vector_data_symmetry_request)["symmetry"]
    assert vector_data_symmetry.shape == (2018, 3)
    assert round(vector_data_symmetry[1009][0], 5) == 0.0023
    assert round(vector_data_symmetry[1009][1], 5) == 1.22311



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _frozen_rotor_simulation:

Impeller-Volute simulation using the Frozen Rotor Approach
----------------------------------------------------------------------
"""

################################################################################################################
# Objective
# ==============================================================================================================
#
# This example demonstrates how to set up a CFD simulation of an impeller and volute using
# the frozen rotor approach. The frozen rotor approach is a technique used in computational
# fluid dynamics (CFD) to simulate the flow around rotating machinery, such as pumps,
# turbines, and compressors. In this approach, the rotor is treated as a stationary
# component, while the surrounding fluid is allowed to move. This allows for a more
# efficient simulation of the flow field, as the rotor's motion does not need to be
# explicitly modeled. The frozen rotor approach is particularly useful for simulating
# steady-state flows, where the rotor's motion is constant.
# By using this approach, engineers can obtain accurate predictions of the
# flow field and performance characteristics of rotating machinery without the need for
# complex and computationally expensive simulations.
#
#
# .. image:: ../../_static/pump-volute-geom.png
#    :align: center
#    :alt: Pump Volute Geometry View

################################################################################################################
# Outline
# ==============================================================================================================

# The example demonstrates the following:

# *Overview & Problem description
# *Launching Fluent in solver mode.
# *Downloading a mesh files from the examples repository.
# *Initial setup
# *Mesh configuration
# *Model selection & Material definition
# *Defining cell zone conditions & boundary conditions
# *Turbomachinery configuration
# *Solver settings
# *Report definitions
# *Initialization
# *Running the simulation
# *Post-processing the results
# *Visualizing the results
# *Saving the case file
# *Closing the solver


################################################################################################################
# Import required libraries/modules
# ==============================================================================================================
import math

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples

################################################################################################################
# Download the mesh file
# ==============================================================================================================

# Download the mesh file from the examples repository
# Impeller and volute mesh files
impeller_mesh = examples.download_file(
    "impeller.msh.h5",
    "pyfluent/examples/pump-volute",
)

volute_mesh = examples.download_file(
    "volute.msh.h5",
    "pyfluent/examples/pump-volute",
)

################################################################################################################
# Define Constants
# ==============================================================================================================

density_water = 998.2  # kg/m^3
viscosity_water = 0.001002  # kg/(m.s)
g = 9.81  # m/s^2
# Impeller speed
impeller_speed = 1450  # rpm
# Convert to rad/s
impeller_speed_rad = impeller_speed * 2 * math.pi / 60  # rad/s

################################################################################################################
# Launch Fluent with solver mode and print Fluent version
# ==============================================================================================================

solver = pyfluent.launch_fluent(
    mode="solver",
    ui_mode="gui",
    processor_count=4,
    cleanup_on_exit=True,
)
print(solver.get_fluent_version())


################################################################################################################
# Read, append the mesh files
# ==============================================================================================================

# Read Impeller mesh file
solver.settings.file.read_mesh(file_name=impeller_mesh)
# Append Volute mesh file
solver.settings.mesh.modify_zones.append_mesh(file_name=volute_mesh)

################################################################################################################
# Display the mesh
# ==============================================================================================================

# Access the graphics object
graphics = solver.settings.results.graphics

# Create a mesh object and configure its settings
mesh_object = solver.settings.results.graphics.mesh.create(name="mesh-1")
mesh_object.surfaces_list = [
    "inlet",
    "mass-flow-inlet-11",
    "interface-impeller-outlet",
    "interface-volute-inlet",
    "blade",
    "impeller-hub",
    "impeller-shroud",
    "inblock-hub",
    "inblock-shroud",
    "volute-inlet-wall",
    "volute-wall",
]
mesh_object.options.edges = True


# Set the hardcopy format for saving the image
graphics.picture.driver_options.hardcopy_format = "png"

graphics.views.restore_view(view_name="front")
mesh_object.display()
graphics.views.auto_scale()
graphics.picture.save_picture(
    file_name="pump-volute-mesh.png",
)


# %%
# .. image:: ../../_static/pump-volute-mesh.png
#    :align: center
#    :width: 650px
#    :height: 400px
#    :alt: Pump Volute Mesh

################################################################################################################
# Set the unit for angular velocity, rad/s to rev/min
# ==============================================================================================================

solver.settings.setup.general.units.set_units(
    quantity="angular-velocity", units_name="rev/min"
)

################################################################################################################
# Define the Viscous Model
# ==============================================================================================================

# Models setting
viscous = solver.settings.setup.models.viscous
viscous = solver.settings.setup.models.viscous
viscous.model = "k-omega"
viscous.k_omega_model = "sst"

################################################################################################################
# Define Materials
# ==============================================================================================================
solver.settings.setup.materials.database.copy_by_name(type="fluid", name="water-liquid")

################################################################################################################
# Define Cell Zone Conditions
# ==============================================================================================================

impeller_cell_zone = solver.settings.setup.cell_zone_conditions.fluid["impeller"]
impeller_cell_zone.general.material = "water-liquid"

impeller_cell_zone.reference_frame.reference_frame_axis_origin = [0, 0, 0]
impeller_cell_zone.reference_frame.reference_frame_axis_direction = [0, 0, 1]
impeller_cell_zone.reference_frame.frame_motion = True

# impeller rotation
impeller_cell_zone.reference_frame.mrf_omega.value = impeller_speed_rad


# Volute Cell Zone Conditions
volute_cell_zone = solver.settings.setup.cell_zone_conditions.fluid["volute"]
volute_cell_zone.general.material = "water-liquid"


# Boundary Conditions
# impeller hub
impeller_hub = solver.settings.setup.boundary_conditions.wall["impeller-hub"].momentum
impeller_hub.wall_motion = "Moving Wall"
impeller_hub.relative = True
impeller_hub.velocity_spec = "Rotational"

#  inblock-shroud

inblock_shroud = solver.settings.setup.boundary_conditions.wall[
    "inblock-shroud"
].momentum
inblock_shroud.wall_motion = "Moving Wall"
inblock_shroud.relative = False
inblock_shroud.velocity_spec = "Rotational"

################################################################################################################
# Define Boundary Conditions
# ==============================================================================================================

# Inlet Boundary Condition
pressure_inlet = solver.settings.setup.boundary_conditions.pressure_inlet["inlet"]
pressure_inlet.momentum.supersonic_or_initial_gauge_pressure.value = -100

# It seems, need to change the boundary condition to mass flow outlet
solver.settings.setup.boundary_conditions.set_zone_type(
    zone_list=["mass-flow-inlet-11"], new_type="mass-flow-outlet"
)
# Outlet Boundary Condition

mass_flow_outlet = solver.settings.setup.boundary_conditions.mass_flow_outlet[
    "mass-flow-inlet-11"
]
mass_flow_outlet.momentum.mass_flow_rate.value = 90  # kg/s


# Create a turbo interfaces
# enable the turbo models
solver.settings.setup.turbo_models.enabled = True

impeller_volute_interface = solver.settings.setup.mesh_interfaces.turbo_create.create(
    adjacent_cell_zone_1="impeller",
    adjacent_cell_zone_2="volute",
    mesh_interface_name="imp-volute-interface",
    turbo_choice="No-Pitch-Scale",
    zone1="interface-impeller-outlet",
    zone2="interface-volute-inlet",
)

################################################################################################################
# Define Solver Settings
# ==============================================================================================================
methods = solver.settings.solution.methods
methods.spatial_discretization.gradient_scheme = "green-gauss-node-based"
methods.high_order_term_relaxation.enable = True

################################################################################################################
# Define Named Expressions
# ==============================================================================================================

pump_head = solver.settings.setup.named_expressions.create("head")
pump_head.definition = "(({p-out} - {p-in}) / (998.2 [kg/m^3] * 9.81[m/s^2]))"
pump_head.output_parameter = True


################################################################################################################
# Define Report Definitions
# ==============================================================================================================

# Create a report definition
# p-out
outlet_pressure_report_def = solver.settings.solution.report_definitions.surface.create(
    "p-out"
)
outlet_pressure_report_def.report_type = "surface-massavg"
outlet_pressure_report_def.surface_names = ["mass-flow-inlet-11"]
outlet_pressure_report_def.field = "total-pressure"
outlet_pressure_report_def.per_surface = False

outlet_pressure_report_plot = solver.settings.solution.monitor.report_plots.create(
    "p-out-rplot"
)
outlet_pressure_report_plot.report_defs = "p-out"

outlet_pressure_report_file = solver.settings.solution.monitor.report_files.create(
    "p-out-rfile"
)
outlet_pressure_report_file.report_defs = "p-out"

# p-in
inlet_pressure_report_def = solver.settings.solution.report_definitions.surface.create(
    "p-in"
)
inlet_pressure_report_def.report_type = "surface-massavg"
inlet_pressure_report_def.surface_names = ["inlet"]
inlet_pressure_report_def.field = "total-pressure"
inlet_pressure_report_def.per_surface = False


# Pump Head
pump_head_report_def = (
    solver.settings.solution.report_definitions.single_valued_expression.create(
        "pump-head"
    )
)

pump_head_report_def.definition = "head"

# report plot
pump_head_report_plot = solver.settings.solution.monitor.report_plots.create(
    "pump-head-rplot"
)
pump_head_report_plot.report_defs = "pump-head"

# report file
pump_head_report_file = solver.settings.solution.monitor.report_files.create(
    "pump-head-rfile"
)
pump_head_report_file.report_defs = "pump-head"

# p-blade
blade_pressure_report_def = solver.settings.solution.report_definitions.surface.create(
    "p-blade"
)
blade_pressure_report_def.report_type = "surface-massavg"
blade_pressure_report_def.surface_names = ["blade"]
blade_pressure_report_def.field = "pressure"
blade_pressure_report_def.per_surface = False


################################################################################################################
# Initialization and run solver
# ==============================================================================================================
initialization = solver.settings.solution.initialization
initialization.reference_frame = "absolute"
initialization.hybrid_init_options.general_settings.initialization_options.initial_pressure = (
    True
)
initialization.hybrid_initialize()

# Run calculation settings
run_calculation = solver.settings.solution.run_calculation
run_calculation.pseudo_time_settings.time_step_method.time_step_size_scale_factor = 10
run_calculation.iter_count = 200

# Write the case file
solver.settings.file.write(file_type="case", file_name="pump_voulte_setup.cas.h5")
# Run the calculation
run_calculation.calculate()

################################################################################################################
# Post-Processing Workflow
# ==============================================================================================================

# Create a mid-plane surface at z = -0.015 m
z_mid_plane = solver.settings.results.surfaces.plane_surface.create(name="z_mid_plane")
z_mid_plane.method = "xy-plane"
z_mid_plane.z = -0.015
z_mid_plane.display()

# Define the contour for static pressure
pressure_contour = solver.settings.results.graphics.contour.create(
    name="static-pressure-contour"
)
pressure_contour.field = "pressure"
pressure_contour.surfaces_list = ["z_mid_plane"]

# Display the contour and save the image

graphics.views.restore_view(view_name="front")
pressure_contour.display()
graphics.views.auto_scale()
graphics.picture.save_picture(file_name="static-pressure-contour.png")


# %%
# .. image:: ../../_static/static-pressure-contour.png
#    :align: center
#    :alt: Static Pressure Contour

################################################################################################################
# Save the case file
# ==============================================================================================================
solver.settings.file.write(
    file_type="case-data", file_name="pump_volute_setup_solved.cas.h5"
)

################################################################################################################
# Close the session
# ==============================================================================================================
solver.exit()

################################################################################################################
# References
# ==============================================================================================================
#
# [1] Ansys Fluent User's Guide, Release 2025R1

# sphinx_gallery_thumbnail_path = '_static/static-pressure-contour.png'



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Session utilities."""

from typing import Any, Dict

import ansys.fluent.core as pyfluent
from ansys.fluent.core.launcher.container_launcher import DockerLauncher
from ansys.fluent.core.launcher.launch_options import (
    Dimension,
    FluentLinuxGraphicsDriver,
    FluentMode,
    FluentWindowsGraphicsDriver,
    Precision,
    UIMode,
)
from ansys.fluent.core.launcher.pim_launcher import PIMLauncher
from ansys.fluent.core.launcher.standalone_launcher import StandaloneLauncher
from ansys.fluent.core.utils.fluent_version import FluentVersion


class SessionBase:
    """Base class for Fluent sessions.

    This class is not intended to be used directly. Instead, use
    the `from_connection`, `from_container`, `from_install`,
    or `from_pim` functions to create a session.
    """

    _session_mode = {
        "Meshing": FluentMode.MESHING,
        "PureMeshing": FluentMode.PURE_MESHING,
        "PrePost": FluentMode.PRE_POST,
        "Solver": FluentMode.SOLVER,
        "SolverAero": FluentMode.SOLVER_AERO,
        "SolverIcing": FluentMode.SOLVER_ICING,
    }

    @classmethod
    def from_install(
        cls,
        ui_mode: UIMode | str | None = None,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        journal_file_names: None | str | list[str] = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        env: Dict[str, Any] = {},  # noqa: B006
        cleanup_on_exit: bool = True,
        dry_run: bool = False,
        start_transcript: bool = True,
        case_file_name: str | None = None,
        case_data_file_name: str | None = None,
        lightweight_mode: bool | None = None,
        py: bool | None = None,
        gpu: bool | None = None,
        cwd: str | None = None,
        fluent_path: str | None = None,
        topy: str | list | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
        certificates_folder: str | None = None,
        insecure_mode: bool = False,
    ):
        """
        Launch a Fluent session in standalone mode.

        Parameters
        ----------
        ui_mode : UIMode
            Defines the user interface mode for Fluent. Options correspond to values in the ``UIMode`` enum.
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version : FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values are from the ``Dimension`` enum
            (``Dimension.TWO`` or ``Dimension.THREE``) or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values are from the ``Precision`` enum (``Precision.SINGLE``
            or ``Precision.DOUBLE``) or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        journal_file_names : str or list of str, optional
            Path(s) to a Fluent journal file(s) that Fluent will execute. Defaults to ``None``.
        start_timeout : int, optional
            Maximum time in seconds allowed for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        env : dict[str, str], optional
            A mapping for modifying environment variables in Fluent. Defaults to ``None``.
        cleanup_on_exit : bool, optional
            Determines whether to shut down the connected Fluent session when exiting PyFluent or calling
            the session's `exit()` method. Defaults to True.
        dry_run : bool, optional
            If True, does not launch Fluent but prints configuration information instead. The `call()` method
            returns a tuple containing the launch string and server info file name. Defaults to False.
        start_transcript : bool, optional
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        case_file_name : str, optional
            Name of the case file to read into the Fluent session. Defaults to None.
        case_data_file_name : str, optional
            Name of the case data file. If both case and data files are provided, they are read into the session.
        lightweight_mode : bool, optional
            If True, runs in lightweight mode where mesh settings are read into a background solver session,
            replacing it once complete. This parameter is only applicable when `case_file_name` is provided; defaults to False.
        py : bool, optional
            If True, runs Fluent in Python mode. Defaults to None.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        cwd : str, optional
            Working directory for the Fluent client.
        fluent_path: str, optional
            User-specified path for Fluent installation.
        topy :  bool or str, optional
            A flag indicating whether to write equivalent Python journals from provided journal files; can also specify
            a filename for the new Python journal.
        start_watchdog : bool, optional
            When `cleanup_on_exit` is True, defaults to True; an independent watchdog process ensures that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any
            Service for uploading/downloading files to/from the server.
        certificates_folder : str, optional
            Path to the folder containing TLS certificates for Fluent's gRPC server.
        insecure_mode : bool, optional
            If True, Fluent's gRPC server will be started in insecure mode without TLS.
            This mode is not recommended. For more details on the implications
            and usage of insecure mode, refer to the Fluent documentation.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        mode = cls._session_mode[cls.__name__]
        argvals = locals().copy()
        argvals.pop("cls", None)  # Remove the class reference from the arguments
        launcher = StandaloneLauncher(**argvals)
        return launcher()

    @classmethod
    def from_container(
        cls,
        ui_mode: UIMode | str | None = None,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        container_dict: dict | None = None,
        dry_run: bool = False,
        cleanup_on_exit: bool = True,
        start_transcript: bool = True,
        py: bool | None = None,
        gpu: bool | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
        certificates_folder: str | None = None,
        insecure_mode: bool = False,
    ):
        """
        Launch a Fluent session in container mode.

        Parameters
        ----------
        ui_mode : UIMode
            Defines the user interface mode for Fluent. Options correspond to values in the ``UIMode`` enum.
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version :  FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            any of ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values include ``Dimension.TWO``,
            ``Dimension.THREE``, or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values include ``Precision.SINGLE``,
            ``Precision.DOUBLE``, or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        start_timeout : int, optional
            Maximum allowable time in seconds for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        container_dict : dict, optional
            Configuration dictionary for launching Fluent inside a Docker container. See also
            :mod:`~ansys.fluent.core.launcher.fluent_container`.
        dry_run : bool, optional
            If True, does not launch Fluent but prints configuration information instead. If dry running a
            container start, this method will return the configured ``container_dict``. Defaults to False.
        cleanup_on_exit : bool
            Determines whether to shut down the connected Fluent session upon exit or when calling
            the session's `exit()` method. Defaults to True.
        start_transcript : bool
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        py : bool, optional
            If True, runs Fluent in Python mode. Defaults to None.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        start_watchdog : bool, optional
            If True and `cleanup_on_exit` is True, an independent watchdog process is run to ensure that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any, optional
            Service for uploading/downloading files to/from the server.
        certificates_folder : str, optional
            Path to the folder containing TLS certificates for Fluent's gRPC server.
        insecure_mode : bool, optional
            If True, Fluent's gRPC server will be started in insecure mode without TLS.
            This mode is not recommended. For more details on the implications
            and usage of insecure mode, refer to the Fluent documentation.

        Returns
        -------
        Meshing | PureMeshing | Solver | SolverIcing | dict
            Session object or configuration dictionary if ``dry_run`` is True.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        mode = cls._session_mode[cls.__name__]
        argvals = locals().copy()
        argvals.pop("cls", None)
        launcher = DockerLauncher(**argvals)
        return launcher()

    @classmethod
    def from_pim(
        cls,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        cleanup_on_exit: bool = True,
        start_transcript: bool = True,
        gpu: bool | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
    ):
        """
        Launch a Fluent session in `PIM <https://pypim.docs.pyansys.com/version/stable/>`_ mode.

        Parameters
        ----------
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version : FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            any of ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values include ``Dimension.TWO``,
            ``Dimension.THREE``, or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values include ``Precision.SINGLE``,
            ``Precision.DOUBLE``, or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        start_timeout : int, optional
            Maximum allowable time in seconds for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        cleanup_on_exit : bool
            Determines whether to shut down the connected Fluent session upon exit or when calling
            the session's `exit()` method. Defaults to True.
        start_transcript : bool
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        start_watchdog : bool, optional
            If True and `cleanup_on_exit` is True, an independent watchdog process is run to ensure that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any, optional
            Service for uploading/downloading files to/from the server.

        Returns
        -------
        Union[Meshing, PureMeshing, Solver, SolverIcing, dict]
            Session object or configuration dictionary if ``dry_run`` is True.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        mode = cls._session_mode[cls.__name__]
        argvals = locals().copy()
        argvals.pop("cls", None)
        launcher = PIMLauncher(**argvals)
        return launcher()

    @classmethod
    def from_connection(
        cls,
        ip: str | None = None,
        port: int | None = None,
        server_info_file_name: str | None = None,
        password: str | None = None,
        allow_remote_host: bool = False,
        certificates_folder: str | None = None,
        insecure_mode: bool = False,
    ):
        """Connect to an existing Fluent server instance.

        Parameters
        ----------
        ip : str, optional
            IP address for connecting to an existing Fluent instance. The
            IP address defaults to ``"127.0.0.1"``. You can also use the environment
            variable ``PYFLUENT_FLUENT_IP=<ip>`` to set this parameter.
            The explicit value of ``ip`` takes precedence over ``PYFLUENT_FLUENT_IP=<ip>``.
        port : int, optional
            Port to listen on for an existing Fluent instance. You can use the
            environment variable ``PYFLUENT_FLUENT_PORT=<port>`` to set a default
            value. The explicit value of ``port`` takes precedence over
            ``PYFLUENT_FLUENT_PORT=<port>``.
        server_info_file_name: str
            Path to server-info file written out by Fluent server. The default is
            ``None``. PyFluent uses the connection information in the file to
            connect to a running Fluent session.
        password : str, optional
            Password to connect to existing Fluent instance.
        allow_remote_host : bool, optional
            Whether to allow connecting to a remote Fluent instance.
        certificates_folder : str, optional
            Path to the folder containing TLS certificates for Fluent's gRPC server.
        insecure_mode : bool, optional
            If True, Fluent's gRPC server will be connected in insecure mode without TLS.
            This mode is not recommended. For more details on the implications
            and usage of insecure mode, refer to the Fluent documentation.

        Raises
        ------
        TypeError
            If the session type does not match the expected session type.
        """
        session = pyfluent.connect_to_fluent(
            ip=ip,
            port=port,
            server_info_file_name=server_info_file_name,
            password=password,
        )

        expected = "Solver" if cls.__name__ == "PrePost" else cls.__name__
        actual = session.__class__.__name__

        if actual != expected:
            raise TypeError(
                f"Session type mismatch: expected {expected}, got {actual}."
            )

        return session


class Meshing(SessionBase):
    """Encapsulates a Fluent server for meshing session connection."""

    pass


class PureMeshing(SessionBase):
    """Encapsulates a Fluent server for pure meshing session connection."""

    pass


class PrePost(SessionBase):
    """Encapsulates a Fluent server for pre-post session connection."""

    pass


class Solver(SessionBase):
    """Encapsulates a Fluent server for solver session connection."""

    pass


class SolverAero(SessionBase):
    """Encapsulates a Fluent server for solver aero session connection."""

    pass


class SolverIcing(SessionBase):
    """Encapsulates a Fluent server for solver icing session connection."""

    pass



"""
Search code for undocumented fields in classes. Requires the ast_comments package.

Usage: python search_undocumented_fields.py ../../src
"""

import os
import sys
from typing import Dict, Set

import ast_comments as ast


def is_ith_undocumented_field(nodes, i: int) -> bool:
    """Check if the ith node is an undocumented field assignment."""
    if i > 0 and isinstance(nodes[i - 1], ast.Comment):
        comment = nodes[i - 1].value.strip()
        print(f"Comment found: {comment}")
        if comment.startswith("#:"):
            return False
    return True


class ClassFieldVisitor(ast.NodeVisitor):
    """AST visitor that extracts public fields from classes."""

    def __init__(self):
        self.classes = {}
        self.current_class = None

    def visit_ClassDef(self, node):
        """Visit a class definition and extract field assignments."""
        old_class = self.current_class
        self.current_class = node.name
        self.classes[node.name] = set()
        if node.name.startswith("_"):
            self.current_class = old_class
            return
        is_enum = False
        for x in node.bases:
            if isinstance(x, ast.Name) and x.id in (
                "Enum",
                "IntEnum",
                "StrEnum",
                "FluentEnum",
            ):
                is_enum = True
            elif isinstance(x, ast.Attribute) and x.attr in (
                "Enum",
                "IntEnum",
                "StrEnum",
                "FluentEnum",
            ):
                is_enum = True
        if is_enum:
            self.current_class = old_class
            return  # Skip Enum classes

        # Visit all child nodes
        for i, child in enumerate(node.body):
            if isinstance(child, ast.Assign):
                # Look for assignments in the class body
                for target in child.targets:
                    if isinstance(target, ast.Name):
                        # Only add public fields (not starting with underscore)
                        if target.id.startswith("_"):
                            continue
                        if is_ith_undocumented_field(node.body, i):
                            self.classes[node.name].add(target.id)
            elif isinstance(child, ast.AnnAssign) and isinstance(
                child.target, ast.Name
            ):
                # Handle annotated assignments (e.g., x: int = 10)
                if child.target.id.startswith("_"):
                    continue
                if is_ith_undocumented_field(node.body, i):
                    self.classes[node.name].add(child.target.id)

        # Visit methods to find instance attribute assignments (self.attr = value)
        for child in node.body:
            if isinstance(child, ast.FunctionDef):
                self.visit(child)

        self.current_class = old_class

    def visit_FunctionDef(self, node):
        """Visit method definitions to find instance attributes."""
        if self.current_class is None:
            return

        # Check if this is an instance method with 'self' parameter
        if node.args.args and node.args.args[0].arg == "self":
            results = list(ast.walk(node))
            for i, child in enumerate(results):
                # Look for self.attr = value assignments
                if isinstance(child, ast.Assign) and any(
                    isinstance(target, ast.Attribute)
                    and isinstance(target.value, ast.Name)
                    and target.value.id == "self"
                    and not target.attr.startswith("_")
                    for target in child.targets
                ):
                    for target in child.targets:
                        if (
                            isinstance(target, ast.Attribute)
                            and isinstance(target.value, ast.Name)
                            and target.value.id == "self"
                        ):
                            if is_ith_undocumented_field(results, i):
                                self.classes[self.current_class].add(target.attr)

                # Handle annotated assignments: self.attr: Type = value
                elif (
                    isinstance(child, ast.AnnAssign)
                    and isinstance(child.target, ast.Attribute)
                    and isinstance(child.target.value, ast.Name)
                    and child.target.value.id == "self"
                    and not child.target.attr.startswith("_")
                ):
                    if is_ith_undocumented_field(results, i):
                        self.classes[self.current_class].add(child.target.attr)


def analyze_file(file_path: str) -> Dict[str, Set[str]]:
    """Analyze a Python file and return classes with their public fields."""
    with open(file_path, "r", encoding="utf-8") as file:
        try:
            tree = ast.parse(file.read(), filename=file_path)
            visitor = ClassFieldVisitor()
            visitor.visit(tree)
            return visitor.classes
        except SyntaxError as e:
            print(f"Syntax error in {file_path}: {e}", file=sys.stderr)
            return {}


def analyze_package(package_path: str) -> Dict[str, Dict[str, Set[str]]]:
    """Analyze all Python files in a package directory."""
    result = {}

    for root, dirs, files in os.walk(package_path):
        # Skip directories named "generated"
        if "generated" in dirs:
            dirs.remove("generated")

        for file in files:
            if file.startswith("_"):
                continue
            if file.endswith(".py"):
                file_path = os.path.join(root, file)
                module_name = os.path.relpath(file_path, package_path).replace(
                    os.path.sep, "."
                )
                module_name = module_name[:-3]  # Remove .py extension

                classes = analyze_file(file_path)
                if classes:
                    result[module_name] = classes

    return result


def write_results(results: Dict[str, Dict[str, Set[str]]], f) -> str:
    """Format the analysis results."""
    for module_name, classes in sorted(results.items()):

        if not classes:
            continue

        module_written = False
        for class_name, fields in sorted(classes.items()):
            if not fields:
                continue

            if not module_written:
                f.write(f"  Module: {module_name}\n")
                module_written = True

            f.write(f"  Class: {class_name}\n")

            for field in sorted(fields):
                f.write(f"    - {field}\n")

        if module_written:
            f.write("\n")


def main():
    """Main function."""
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <package_directory>", file=sys.stderr)
        sys.exit(1)

    package_path = sys.argv[1]
    if not os.path.isdir(package_path):
        print(f"Error: {package_path} is not a directory", file=sys.stderr)
        sys.exit(1)

    results = analyze_package(package_path)
    with open("undocumented_fields.txt", "w", encoding="utf-8") as f:
        write_results(results, f)


if __name__ == "__main__":
    main()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
from util.solver import copy_database_material


@pytest.mark.settings_only
@pytest.mark.fluent_version("latest")
def test_solver_material(mixing_elbow_settings_session):
    solver_session = mixing_elbow_settings_session
    setup_materials = solver_session.setup.materials
    copy_database_material(materials=setup_materials, type="fluid", name="water-liquid")
    elbow_fluid = solver_session.setup.cell_zone_conditions.fluid["elbow-fluid"]

    assert "water-liquid" not in elbow_fluid.material()
    elbow_fluid.material = "water-liquid"
    assert "water-liquid" in elbow_fluid.material()

    setup_materials.database.copy_by_name(type="fluid", name="air")
    elbow_fluid.material = "air"
    assert "air" in elbow_fluid.material()
    assert "air" in elbow_fluid.material()

    assert setup_materials.child_names == [
        "database",
        "fluid",
        "solid",
        "mixture",
        "inert_particle",
        "droplet_particle",
        "combusting_particle",
        "particle_mixture",
    ]
    assert setup_materials.database.get_active_command_names() == [
        "copy_by_formula",
        "copy_by_name",
        "list_materials",
        "list_properties",
    ]
    setup_materials.database.copy_by_formula(type="fluid", formula="c2h6")
    elbow_fluid.material = "ethane"
    assert "ethane" in elbow_fluid.material()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ahmed_body_simulation:

Ahmed Body External Aerodynamics Simulation
-------------------------------------------
"""

#######################################################################################
# Objective
# =====================================================================================
#
# Ahmed body is a simplified car model used for studying the flow around it and to
# predict the drag and lift forces. The model consists of a slanted back and a blunt
# front.
#
# In this example, PyFluent API is used to perform Ahmed Body external aerodynamics
# simulation. which includes typical workflow of CFD Simulation as follows:
#
# * Importing the geometry/CAD model.
# * Meshing the geometry.
# * Setting up the solver.
# * Running the solver.
# * Post-processing the results.
#
#
#
# .. image:: ../../_static/ahmed_body_model.png
#    :align: center
#    :alt: Ahmed Body Model


#######################################################################################
# Import required libraries/modules
# =====================================================================================

import platform

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples
from ansys.fluent.visualization import Contour, GraphicsWindow, set_config

#######################################################################################
# Configure specific settings for this example
# =====================================================================================
set_config(blocking=True, set_view_on_display="isometric")

#######################################################################################
# Launch Fluent session with meshing mode and print Fluent version
# =====================================================================================
session = pyfluent.launch_fluent(mode="meshing", cleanup_on_exit=True)
print(session.get_fluent_version())

#######################################################################################
# Meshing Workflow
# =====================================================================================

#######################################################################################
# Initialize the Meshing Workflow
# =====================================================================================

workflow = session.workflow

filenames = {
    "Windows": "ahmed_body_20_0degree_boi_half.scdoc",
    "Other": "ahmed_body_20_0degree_boi_half.scdoc.pmdb",
}

geometry_filename = examples.download_file(
    filenames.get(platform.system(), filenames["Other"]),
    "pyfluent/examples/Ahmed-Body-Simulation",
)

workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
workflow.TaskObject["Import Geometry"].Arguments = dict(FileName=geometry_filename)
workflow.TaskObject["Import Geometry"].Execute()

#######################################################################################
# Add Local Face Sizing
# =====================================================================================
add_local_sizing = workflow.TaskObject["Add Local Sizing"]
add_local_sizing.Arguments = dict(
    {
        "AddChild": "yes",
        "BOIControlName": "facesize_front",
        "BOIFaceLabelList": ["wall_ahmed_body_front"],
        "BOIGrowthRate": 1.15,
        "BOISize": 8,
    }
)
add_local_sizing.Execute()

add_local_sizing.InsertCompoundChildTask()
workflow.TaskObject["Add Local Sizing"].Execute()
add_local_sizing = workflow.TaskObject["Add Local Sizing"]
add_local_sizing.Arguments = dict(
    {
        "AddChild": "yes",
        "BOIControlName": "facesize_rear",
        "BOIFaceLabelList": ["wall_ahmed_body_rear"],
        "BOIGrowthRate": 1.15,
        "BOISize": 5,
    }
)
add_local_sizing.Execute()

add_local_sizing.InsertCompoundChildTask()
workflow.TaskObject["Add Local Sizing"].Execute()
add_local_sizing = workflow.TaskObject["Add Local Sizing"]
add_local_sizing.Arguments = dict(
    {
        "AddChild": "yes",
        "BOIControlName": "facesize_main",
        "BOIFaceLabelList": ["wall_ahmed_body_main"],
        "BOIGrowthRate": 1.15,
        "BOISize": 12,
    }
)
add_local_sizing.Execute()

#######################################################################################
# Add BOI (Body of Influence) Sizing
# =====================================================================================
add_boi_sizing = workflow.TaskObject["Add Local Sizing"]
add_boi_sizing.InsertCompoundChildTask()
add_boi_sizing.Arguments = dict(
    {
        "AddChild": "yes",
        "BOIControlName": "boi_1",
        "BOIExecution": "Body Of Influence",
        "BOIFaceLabelList": ["ahmed_body_20_0degree_boi_half-boi"],
        "BOISize": 20,
    }
)
add_boi_sizing.Execute()
add_boi_sizing.InsertCompoundChildTask()


#######################################################################################
# Add Surface Mesh Sizing
# =====================================================================================
generate_surface_mesh = workflow.TaskObject["Generate the Surface Mesh"]
generate_surface_mesh.Arguments = dict(
    {
        "CFDSurfaceMeshControls": {
            "CurvatureNormalAngle": 12,
            "GrowthRate": 1.15,
            "MaxSize": 50,
            "MinSize": 1,
            "SizeFunctions": "Curvature",
        }
    }
)

generate_surface_mesh.Execute()
generate_surface_mesh.InsertNextTask(CommandName="ImproveSurfaceMesh")
improve_surface_mesh = workflow.TaskObject["Improve Surface Mesh"]
improve_surface_mesh.Arguments.update_dict({"FaceQualityLimit": 0.4})
improve_surface_mesh.Execute()

#######################################################################################
# Describe Geometry, Update Boundaries, Update Regions
# =====================================================================================
workflow.TaskObject["Describe Geometry"].Arguments = dict(
    CappingRequired="Yes",
    SetupType="The geometry consists of only fluid regions with no voids",
)
workflow.TaskObject["Describe Geometry"].Execute()
workflow.TaskObject["Update Boundaries"].Execute()
workflow.TaskObject["Update Regions"].Execute()

#######################################################################################
# Add Boundary Layers
# =====================================================================================
add_boundary_layers = workflow.TaskObject["Add Boundary Layers"]
add_boundary_layers.AddChildToTask()
add_boundary_layers.InsertCompoundChildTask()
workflow.TaskObject["smooth-transition_1"].Arguments.update_dict(
    {
        "BLControlName": "smooth-transition_1",
        "NumberOfLayers": 14,
        "Rate": 1.15,
        "TransitionRatio": 0.5,
    }
)
add_boundary_layers.Execute()


#######################################################################################
# Generate the Volume Mesh
# =====================================================================================
generate_volume_mesh = workflow.TaskObject["Generate the Volume Mesh"]
generate_volume_mesh.Arguments.update_dict({"VolumeFill": "poly-hexcore"})
generate_volume_mesh.Execute()

#######################################################################################
# Switch to the Solver Mode
# =====================================================================================
session = session.switch_to_solver()

#######################################################################################
# Mesh Visualization
# =====================================================================================

# %%
# .. image:: ../../_static/ahmed_body_mesh_1.png
#    :align: center
#    :alt: Ahmed Body Mesh

# %%
# .. image:: ../../_static/ahmed_body_mesh_2.png
#    :align: center
#    :alt: Ahmed Body Mesh

#######################################################################################
# Solver Setup and Solve Workflow
# =====================================================================================

#######################################################################################
# Define Constants
# =====================================================================================
density = 1.225
inlet_velocity = 30
inlet_area = 0.11203202

#######################################################################################
# Define Materials
# =====================================================================================
session.tui.define.materials.change_create("air", "air", "yes", "constant", density)
session.settings.setup.models.viscous.model = "k-epsilon"
session.settings.setup.models.viscous.k_epsilon_model = "realizable"
session.settings.setup.models.viscous.options.curvature_correction = True

#######################################################################################
# Define Boundary Conditions
# =====================================================================================
inlet = session.settings.setup.boundary_conditions.velocity_inlet["inlet"]
inlet.turbulence.turb_intensity = 0.05
inlet.momentum.velocity.value = inlet_velocity
inlet.turbulence.turb_viscosity_ratio = 5

outlet = session.settings.setup.boundary_conditions.pressure_outlet["outlet"]
outlet.turbulence.turb_intensity = 0.05

#######################################################################################
# Define Reference Values
# =====================================================================================
session.settings.setup.reference_values.area = inlet_area
session.settings.setup.reference_values.density = density
session.settings.setup.reference_values.velocity = inlet_velocity

#######################################################################################
# Define Solver Settings
# =====================================================================================
session.tui.solve.set.p_v_coupling(24)

session.tui.solve.set.discretization_scheme("pressure", 12)
session.tui.solve.set.discretization_scheme("k", 1)
session.tui.solve.set.discretization_scheme("epsilon", 1)
session.tui.solve.initialize.set_defaults("k", 0.000001)

session.settings.solution.monitor.residual.equations["continuity"].absolute_criteria = (
    0.0001
)
session.settings.solution.monitor.residual.equations["x-velocity"].absolute_criteria = (
    0.0001
)
session.settings.solution.monitor.residual.equations["y-velocity"].absolute_criteria = (
    0.0001
)
session.settings.solution.monitor.residual.equations["z-velocity"].absolute_criteria = (
    0.0001
)
session.settings.solution.monitor.residual.equations["k"].absolute_criteria = 0.0001
session.settings.solution.monitor.residual.equations["epsilon"].absolute_criteria = (
    0.0001
)

#######################################################################################
# Define Report Definitions
# =====================================================================================

session.settings.solution.report_definitions.drag["cd-mon1"] = {}
session.settings.solution.report_definitions.drag["cd-mon1"] = {
    "zones": ["wall_ahmed_body_main", "wall_ahmed_body_front", "wall_ahmed_body_rear"],
    "force_vector": [0, 0, 1],
}
session.parameters.output_parameters.report_definitions.create(name="parameter-1")
session.parameters.output_parameters.report_definitions["parameter-1"] = {
    "report_definition": "cd-mon1"
}

session.settings.solution.monitor.report_plots.create(name="cd-mon1")
session.settings.solution.monitor.report_plots["cd-mon1"] = {"report_defs": ["cd-mon1"]}

#######################################################################################
# Initialize and Run Solver
# =====================================================================================

session.settings.solution.run_calculation.iter_count = 5
session.settings.solution.initialization.initialization_type = "standard"
session.settings.solution.initialization.standard_initialize()
session.settings.solution.run_calculation.iterate(iter_count=5)

#######################################################################################
# Post-Processing Workflow
# =====================================================================================
session.results.surfaces.iso_surface.create(name="xmid")
session.results.surfaces.iso_surface["xmid"].field = "x-coordinate"
session.results.surfaces.iso_surface["xmid"] = {"iso_values": [0]}

contour1 = Contour(solver=session, field="velocity-magnitude", surfaces=["xmid"])
disp1 = GraphicsWindow()
disp1.add_graphics(contour1)
disp1.show()

contour2 = Contour(solver=session, surfaces=["xmid"])
contour2.field.allowed_values
contour2.field = "pressure-coefficient"
disp2 = GraphicsWindow()
disp2.add_graphics(contour2)
disp2.show()

#######################################################################################
# Simulation Results Visualization
# =====================================================================================

# %%
# .. image:: ../../_static/ahmed_body_model_velocity_mag.png
#    :align: center
#    :alt: Velocity Magnitude

# %%
#    Velocity Magnitude Contour

# %%
# .. image:: ../../_static/ahmed_body_model_pressure_coeff.png
#    :align: center
#    :alt: Peressure Coefficient

# %%
#    Pressure Coefficient Contour

#######################################################################################
# Save the case file
# =====================================================================================
session.settings.file.write(file_type="case-data", file_name="ahmed_body_final.cas.h5")

#######################################################################################
# Close the session
# =====================================================================================
session.exit()


#######################################################################################
# References
# =====================================================================================
#
# [1] S.R. Ahmed, G. Ramm, Some Salient Features of the Time-Averaged Ground Vehicle
# Wake,SAE-Paper 840300,1984

# sphinx_gallery_thumbnail_path = '_static/ahmed_body_model_pressure_coeff.png'



"""Provide a module to generate the documentation classes for Fluent settings tree.
Running this module generates a .rst files for the Fluent.

settings classes. The out is placed at:
- doc/source/api/solver/_autosummary/settings
Process
-------
    - From the settings API classes recursively generate the list of parents for the current class.
    -- Populate a parents dictionary with current class file name (not class name) as key and list of parents file names (not class names) as value.
    - Recursively Generate the rst files for classes starting with settings.root.
    -- Add target reference as the file name for the given class. This is used by other classes to generate hyperlinks
    -- Add properties like members, undoc-memebers, show-inheritence to the autoclass directive.
    -- Generate the tables of children, commands, arguments, and parents.
    --- Get access to the respective properties and members on the class with get_attr.
    --- Use the file name of the child class to generate the hyperlink to that class.
    --- Use the __doc__ property to generate the short summary for the corresponding child
    --- Use the previously generated perents dict to populate the parents table.
Usage
-----
python <path to settings_rstgen.py>
"""

from contextlib import redirect_stdout
import importlib
import io
import os
from pathlib import Path

from deprecated_pyfluent_apis import PYFLUENT_DEPRECATED_DATA

from ansys.fluent.core.search import search
from ansys.fluent.core.utils.fluent_version import (
    FluentVersion,
    get_version_for_file_name,
)

parents_dict = {}
rst_list = []
deprecated_class_version = {}


def _get_indent_str(indent):
    return f"{' '*indent*4}"


def _generate_table_for_rst(r, data_dict=None):
    if data_dict is None:
        data_dict = {}
    # Get dimensions for columns
    key_max = len(max(data_dict.keys(), key=len))
    val_max = len(max(data_dict.values(), key=len))
    col_gap = 3
    total = key_max + val_max + col_gap
    # Top border
    r.write(f'{"="*key_max}{" "*col_gap}{"="*val_max}\n\n')
    header = True
    for key, value in data_dict.items():
        if header:
            # Write header and border
            r.write(f'{key}{" "*(total-len(key)-len(value))}{value}\n\n')
            r.write(f'{"="*key_max}{" "*col_gap}{"="*val_max}\n')
            header = False
        else:
            # actual data
            r.write(f'{key}{" "*(total-len(key)-len(value))}{value}\n\n')
    # Bottom border
    r.write(f'{"="*key_max}{" "*col_gap}{"="*val_max}\n\n')


def _populate_parents_list(cls):
    if hasattr(cls, "child_names"):
        for child in cls.child_names:
            child_cls = cls._child_classes[child]
            child_cls_name = child_cls.__name__
            if not parents_dict.get(child_cls_name):
                parents_dict[child_cls_name] = []
            if cls not in parents_dict[child_cls_name]:
                parents_dict[child_cls_name].append(cls)

    if hasattr(cls, "command_names"):
        for child in cls.command_names:
            child_cls = cls._child_classes[child]
            child_cls_name = child_cls.__name__
            if not parents_dict.get(child_cls_name):
                parents_dict[child_cls_name] = []
            if cls not in parents_dict[child_cls_name]:
                parents_dict[child_cls_name].append(cls)

    if hasattr(cls, "argument_names"):
        for child in cls.argument_names:
            child_cls = cls._child_classes[child]
            child_cls_name = child_cls.__name__
            if not parents_dict.get(child_cls_name):
                parents_dict[child_cls_name] = []
            if cls not in parents_dict[child_cls_name]:
                parents_dict[child_cls_name].append(cls)

    if hasattr(cls, "child_object_type"):
        child_cls = getattr(cls, "child_object_type")
        child_cls_name = child_cls.__name__
        if not parents_dict.get(child_cls_name):
            parents_dict[child_cls_name] = []
        if cls not in parents_dict[child_cls_name]:
            parents_dict[child_cls_name].append(cls)

    if hasattr(cls, "child_names"):
        for child in cls.child_names:
            _populate_parents_list(cls._child_classes[child])

    if hasattr(cls, "command_names"):
        for child in cls.command_names:
            _populate_parents_list(cls._child_classes[child])

    if hasattr(cls, "argument_names"):
        for child in cls.argument_names:
            _populate_parents_list(cls._child_classes[child])

    if hasattr(cls, "child_object_type"):
        _populate_parents_list(getattr(cls, "child_object_type"))


def _populate_rst_from_settings(rst_dir, cls, version):
    istr1 = _get_indent_str(1)
    cls_name = cls.__name__
    cls_orig_name = cls._python_name
    rstpath = os.path.normpath(os.path.join(rst_dir, cls_name + ".rst"))
    has_children = hasattr(cls, "child_names") and len(cls.child_names) > 0
    has_commands = hasattr(cls, "command_names") and len(cls.command_names) > 0
    has_arguments = hasattr(cls, "argument_names") and len(cls.argument_names) > 0
    has_named_object = hasattr(cls, "child_object_type")
    with open(rstpath, "w") as r:
        # Populate initial rst
        r.write(":orphan:\n\n")
        if cls_orig_name == "root":
            r.write(f".. _ref_{cls_name}:\n\n")
        else:
            r.write(f".. _{cls_name}:\n\n")
        r.write(f"{cls_orig_name}\n")
        r.write(f'{"="*(len(cls_orig_name))}\n\n')
        deprecated = getattr(cls, "_deprecated_version", None)
        if deprecated:
            r.write(f".. deprecated:: Ansys {cls._deprecated_version}\n\n")
            deprecated_class_version.update({cls_name: deprecated})
        r.write(
            f".. autoclass:: ansys.fluent.core.generated.solver.settings_{version}.{cls_name}\n"
        )
        r.write(f"{istr1}:show-inheritance:\n\n")

        if has_children:
            r.write(".. rubric:: Attributes\n\n")
            data_dict = {}
            data_dict["Attribute"] = "Summary"
            for child in cls.child_names:
                child_cls = cls._child_classes[child]
                ref_string = f":ref:`{child} <{child_cls.__name__}>`"
                data_dict[ref_string] = child_cls.__doc__.strip("\n").split("\n")[0]
            _generate_table_for_rst(r, data_dict)

        if has_commands:
            r.write(".. rubric:: Methods\n\n")
            data_dict = {}
            data_dict["Method"] = "Summary"
            for child in cls.command_names:
                child_cls = cls._child_classes[child]
                ref_string = f":ref:`{child} <{child_cls.__name__}>`"
                data_dict[ref_string] = child_cls.__doc__.strip("\n").split("\n")[0]
            _generate_table_for_rst(r, data_dict)

        if has_arguments:
            r.write(".. rubric:: Arguments\n\n")
            data_dict = {}
            data_dict["Argument"] = "Summary"
            for child in cls.argument_names:
                child_cls = cls._child_classes[child]
                ref_string = f":ref:`{child} <{child_cls.__name__}>`"
                data_dict[ref_string] = child_cls.__doc__.strip("\n").split("\n")[0]
            _generate_table_for_rst(r, data_dict)

        if has_named_object:
            child_cls = getattr(cls, "child_object_type")
            ref_string = f":ref:`{child_cls.__name__} <{child_cls.__name__}>`"
            data_dict = {}
            data_dict[ref_string] = child_cls.__doc__.strip("\n").split("\n")[0]
            r.write(".. rubric:: Named object type\n\n")
            r.write(f"{ref_string}\n\n\n")

        if parents_dict.get(cls_name):
            r.write(".. rubric:: Included in:\n\n")
            data_dict = {}
            data_dict["Parent"] = "Summary"
            for parent in parents_dict.get(cls_name):
                parent_ref = parent.__name__
                if parent_ref == "root":
                    parent_ref = "ref_root"
                ref_string = f":ref:`{parent.__name__} <{parent_ref}>`"
                data_dict[ref_string] = parent.__doc__.strip("\n").split("\n")[0]
            _generate_table_for_rst(r, data_dict)

    if rstpath not in rst_list:
        rst_list.append(rstpath)
        if has_children:
            for child in cls.child_names:
                _populate_rst_from_settings(rst_dir, cls._child_classes[child], version)

        if has_commands:
            for child in cls.command_names:
                _populate_rst_from_settings(rst_dir, cls._child_classes[child], version)

        if has_arguments:
            for child in cls.argument_names:
                _populate_rst_from_settings(rst_dir, cls._child_classes[child], version)

        if has_named_object:
            _populate_rst_from_settings(
                rst_dir, getattr(cls, "child_object_type"), version
            )


def _write_deprecated_rst_table(rst_dir, deprecated_class_version):
    deprecated_rst = (Path(rst_dir).parents[2] / "deprecated_apis.rst").resolve()
    if deprecated_rst.exists():
        deprecated_rst.unlink()
    else:
        deprecated_rst.touch()

    deprecated_data = []
    fluent_header = ["Target", "Deprecated"]
    pyflunet_header = ["Target", "Deprecated", "Alternatives"]
    name = "Deprecated APIs"
    pyfluent_name = "Deprecated PyFluent APIs"
    fluent_name = "Deprecated Ansys Fluent APIs"
    buffer = io.StringIO()

    for class_name, deprecated_version in deprecated_class_version.items():
        with redirect_stdout(buffer):
            search(class_name)
        output = buffer.getvalue()
        out = output.split("\n")
        settings = set(
            [
                setting
                for setting in out
                if "tui" not in setting and "meshing" not in setting
            ]
        )
        for setting in settings:
            if setting and setting.split(".")[-1].split(" ")[0] == class_name:
                setting = (
                    setting.replace("<solver_session>", "solver")
                    .replace("<", "")
                    .replace(">", "")
                )
                settings_with_ref = f":ref:`{setting} <{class_name}>`"
                deprecated_data.append((settings_with_ref, deprecated_version))

    with open(deprecated_rst, "w", encoding="utf-8") as f:
        f.write(":orphan:\n\n")
        f.write(f"{name}\n")
        f.write(f'{"="*(len(name))}\n\n')

        f.write(f"{pyfluent_name}\n")
        f.write(f'{"-"*(len(pyfluent_name))}\n\n')
        f.write(".. list-table:: Deprecated PyFluent APIs\n")
        f.write("   :header-rows: 1\n\n")
        f.write("   * - " + "\n     - ".join(pyflunet_header) + "\n")
        for row in PYFLUENT_DEPRECATED_DATA:
            f.write("   * - " + "\n     - ".join(row) + "\n")

        f.write(f"{fluent_name}\n")
        f.write(f'{"-"*(len(fluent_name))}\n\n')
        f.write(".. list-table:: Deprecated Ansys Fluent APIs\n")
        f.write("   :header-rows: 1\n\n")
        f.write("   * - " + "\n     - ".join(fluent_header) + "\n")
        sorted_data = sorted(deprecated_data, key=lambda x: len(x[0]))
        for row in sorted_data:
            f.write("   * - " + "\n     - ".join(row) + "\n")


if __name__ == "__main__":
    print("Generating rst files for settings API classes")
    dirname = os.path.dirname(__file__)
    rst_dir = os.path.normpath(
        os.path.join(
            dirname,
            "source",
            "api",
            "solver",
            "_autosummary",
            "settings",
        )
    )

    if not os.path.exists(rst_dir):
        os.makedirs(rst_dir)

    image_tag = os.getenv(
        "FLUENT_IMAGE_TAG", FluentVersion.current_dev().docker_image_tag
    )
    version = get_version_for_file_name(image_tag.lstrip("v"))
    settings = importlib.import_module(
        f"ansys.fluent.core.generated.solver.settings_{version}"
    )
    _populate_parents_list(settings.root)
    _populate_rst_from_settings(rst_dir, settings.root, version)
    _write_deprecated_rst_table(rst_dir, deprecated_class_version)
